/* $Id: dpd.c,v 1.16 2004/12/15 07:36:32 kamada Exp $ */
/*
 * Copyright (C) 2003, 2004 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "../lib/vmbuf.h"
#include "../lib/rc_type.h"
#include "../lib/rc_net.h"
#include "utils.h"
#include "sockmisc.h"
#include "handle.h"
#include "peer.h"
#include "scheduler.h"
#include "pfkey.h"
#include "dpd.h"


static void dpd_expunge_dead1(struct kink_handle *kh);


/*
 * Register dead peer and create new kink_peer.
 * <kh->peer> is usually equal to <peer>.
 */
struct kink_peer *
dpd_peer_recreate(struct kink_handle *kh,
    struct kink_peer *peer, uint32_t epoch)
{
	struct kink_peer *newp;

	kinkd_log(KLLV_WARNING,
	    "dead peer queued (p=%s, old epoch=%u, new epoch=%u)\n",
	    peer->remote_principal, peer->epoch, epoch);

	newp = kink_peer_dup_shallow(kh, peer);
	newp->epoch = epoch;

	peer->dead = 1;
	SLIST_INSERT_HEAD(&kh->g->deadpeerlist, peer, next);

	return newp;
}

/*
 * Called by main loop, and expunge all SAs with dead peers.
 */
void
dpd_expunge_dead(struct kink_global *kg)
{
	struct kink_peer *dead;
	struct kink_handle *kh;

	while ((dead = SLIST_FIRST(&kg->deadpeerlist)) != NULL) {
		SLIST_REMOVE_HEAD(&kg->deadpeerlist, next);

		kinkd_log(KLLV_INFO,
		    "expunging old SAs with dead peer (p=%s)\n",
		    dead->remote_principal);
		while ((kh = sched_get_by_peer(kg, dead)) != NULL) {
			dpd_expunge_dead1(kh);
			sched_unregister(kh);
			if (kh->state->cancel != NULL)
				(*kh->state->cancel)(kh);
			/* release kind_handle */
			if (kh->ph2 != NULL)
				release_ph2(kh->ph2);
			release_handle(kh);
		}

		free(dead);
	}
}

static void
dpd_expunge_dead1(struct kink_handle *kh)
{
	int ret;

	if (kh->ph2 == NULL || kh->ph2->approval == NULL) {
		kinkd_log(KLLV_INFO,
		    "expunge (state=%s)\n", kh->state->strname);
		return;
	}

	kinkd_log(KLLV_INFO,
	    "expunge (state=%s, %s --> %s)\n",
	    kh->state->strname,
	    rcs_sa2str(kh->ph2->src), rcs_sa2str(kh->ph2->dst));

	/* delete outbound SA */
	ret = pk_senddelete(kh->g->fd_pfkey, kh->ph2->approval,
	    kh->ph2->src, kh->ph2->dst, RCT_DIR_OUTBOUND);
	if (ret != 0) {
		kinkd_log(KLLV_ERR, "failed to pk_senddelete\n");
		return;
	}
	/* XXX delete inbound SA or not? */
}
