/* $Id: handle.c,v 1.25 2004/12/09 02:29:41 kamada Exp $ */
/*
 * Copyright (C) 2003, 2004 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "config.h"

#include <sys/types.h>

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>

#include "../lib/vmbuf.h"
#include "utils.h"
#include "bbkk.h"
#include "proposal.h"
#include "handle.h"


const struct kink_state state_none = {
	1, "NONE",
	NULL, 0,
	NULL, NULL
};


/*
 * kink_handle allocatoin/deallocation
 */
struct kink_handle *
allocate_handle(struct kink_global *kg)
{
	static const struct kink_handle kh0;
	struct kink_handle *kh;

	if ((kh = (struct kink_handle *)malloc(sizeof(*kh))) == NULL) {
		kinkd_log(KLLV_FATAL, "out of memory\n");
		return NULL;
	}

	*kh = kh0;
	kh->g = kg;
	kh->state = &state_none;

	return kh;
}

void
release_handle(struct kink_handle *kh)
{
	vfree(kh->in_isakmp);
	vfree(kh->krb_ap_req);
	vfree(kh->krb_ap_rep);
	if (kh->remote_sa != NULL)
		free(kh->remote_sa);
	vfree(kh->cache_reply);

	release_payloads(kh);
	release_auth_contexts(kh);

	if (kh->ka != NULL)
		kh->ka->refcnt--;

	free(kh);
}

void
release_payloads(struct kink_handle *kh)
{
	vfree(kh->ap_req);
	vfree(kh->ap_rep);
	vfree(kh->krb_error);
	vfree(kh->isakmp);
	vfree(kh->encrypt);
	vfree(kh->error);

	kh->ap_req = NULL;
	kh->ap_rep = NULL;
	kh->krb_error = NULL;
	kh->isakmp = NULL;
	kh->encrypt = NULL;
	kh->error = NULL;
}

/*
 * Caution: This function does not care for single auth_context.
 */
void
release_auth_contexts(struct kink_handle *kh)
{
	while (kh->v_auth_context_num > 0)
		bbkk_free_auth_context(kh->g->context,
		    kh->v_auth_contexts[--kh->v_auth_context_num]);
	if (kh->auth_context_ack != NULL) {
		bbkk_free_auth_context(kh->g->context, kh->auth_context_ack);
		kh->auth_context_ack = NULL;
	}
}

/*
 * ph2handle allocator/deallocator
 */

struct ph2handle *
allocate_ph2(int side)
{
	static struct ph2handle ph2_0;
	struct ph2handle *ph2;

	if ((ph2 = (struct ph2handle *)malloc(sizeof(*ph2))) == NULL) {
		kinkd_log(KLLV_FATAL, "out of memory\n");
		return NULL;
	}

	*ph2 = ph2_0;
	ph2->side = side;

	return ph2;
}

void
release_ph2(struct ph2handle *ph2)
{
	if (ph2->src != NULL)
		free(ph2->src);
	if (ph2->dst != NULL)
		free(ph2->dst);

	flushsaprop(ph2->proposal);
	flushsaprop(ph2->approval);

	vfree(ph2->sa);
	vfree(ph2->nonce);
	vfree(ph2->nonce_p);
	vfree(ph2->id);
	vfree(ph2->id_p);
	vfree(ph2->dhpub_p);
	vfree(ph2->sa_ret);

	free(ph2);
}
