/* $Id: cfsetup.c,v 1.43 2005/01/05 05:57:41 sakane Exp $ */
/*
 * Copyright (C) 2004 WIDE Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/param.h>

#include <stdlib.h>
#include <netdb.h>		/* EAI_NONAME */
#include <string.h>
#include <stdio.h>

#include "racoon.h"
#include "cfsetup.h"

#define RCF_CALL_TDF(cl, dd) \
do { \
	int (*func)(); \
	if (rcf_check_cft((cl), CFT_DIRECTIVE)) \
		goto err; \
	if ((func = rcf_get_tdf((cl)->d.dir)) == 0) { \
		plog(PLOG_INTERR, PLOGLOC, NULL, \
		    "error the function for %d found in %s %d\n", \
			(cl)->d.dir, (cl)->file, (cl)->lineno); \
		goto err; \
	} \
	if ((*func)((cl), (dd))) { \
		plog(PLOG_INTERR, PLOGLOC, NULL, \
		    "error in %s %d\n", (cl)->file, (cl)->lineno); \
		goto err; \
	} \
} while(0)

#define DEEPCOPY_VDUP(src, dst) \
do { \
	if ((src) != 0 && ((dst) = vdup(src)) == 0) \
		goto err; \
} while(0)

#define DEEPCOPY_KMP(src, dst) \
do { \
	if ((src) != 0 && ((dst) = rcf_deepcopy_kmp(src)) == 0) \
		goto err; \
} while(0)

#define DEEPCOPY_PKLIST(src, dst) \
do { \
	if ((src) != 0 && ((dst) = rcf_deepcopy_pklist(src)) == 0) \
		goto err; \
} while(0)

#define DEEPCOPY_ADDRLIST(src, dst) \
do { \
	if ((src) != 0 && ((dst) = rcf_deepcopy_addrlist(src)) == 0) \
		goto err; \
} while(0)

#define DEEPCOPY_IDLIST(src, dst) \
do { \
	if ((src) != 0 && ((dst) = rcf_deepcopy_idlist(src)) == 0) \
		goto err; \
} while(0)

#define DEEPCOPY_ALGLIST(src, dst) \
do { \
	if ((src) != 0 && ((dst) = rcf_deepcopy_alglist(src)) == 0) \
		goto err; \
} while(0)

#define DEEPCOPY_POLICY(src, dst) \
do { \
	if ((src) != 0 && ((dst) = rcf_deepcopy_policy(src)) == 0) \
		goto err; \
} while(0)

#define DEEPCOPY_IPSEC(src, dst) \
do { \
	if ((src) != 0 && ((dst) = rcf_deepcopy_ipsec(src)) == 0) \
		goto err; \
} while(0)

#define DEEPCOPY_SA(src, dst) \
do { \
	if ((src) != 0 && ((dst) = rcf_deepcopy_sa(src)) == 0) \
		goto err; \
} while(0)

#define DEEPCOPY_LOG(src, dst) \
do { \
	if ((src) != 0 && ((dst) = rcf_deepcopy_log(src)) == 0) \
		goto err; \
} while(0)

extern struct cf_lists *cf_lists;

struct rcf_setval *rcf_setval_head = 0;
struct rcf_default *rcf_default_head = 0;
struct rcf_interface *rcf_interface_head = 0;
struct rcf_resolver *rcf_resolver_head = 0;
struct rcf_remote *rcf_remote_head = 0;
struct rcf_selector *rcf_selector_head = 0;

static int (*rcf_get_tdf __P((rcf_tdir)))();
	/* setval */
static int rcf_fix_setval __P((struct rcf_setval **));
static void rcf_clean_setval_list __P((struct rcf_setval *));
static void rcf_clean_setval __P((struct rcf_setval *));
	/* interface */
static int rcf_fix_interface __P((struct rcf_interface **));
static void rcf_clean_interface_list __P((struct rcf_interface *));
static int rcf_fix_if_ike __P((struct cf_list *, void *));
static int rcf_fix_if_kink __P((struct cf_list *, void *));
static int rcf_fix_if_spmd __P((struct cf_list *, void *));
	/* resolver */
static int rcf_fix_resolver __P((struct rcf_resolver **));
static void rcf_clean_resolver_list __P((struct rcf_resolver *));
static int rcf_fix_nameserver __P((struct cf_list *, void *));
static int rcf_fix_dns_query __P((struct cf_list *, void *));
static int rcf_fix_resolver_enable __P((struct cf_list *, void *));
	/* remote */
static int rcf_fix_remote __P((struct rcf_remote **));
static void rcf_clean_remote_list __P((struct rcf_remote *));
static void rcf_clean_remote __P((struct rcf_remote *));
static struct rcf_remote *rcf_deepcopy_remote __P((struct rcf_remote *));
static int rcf_fix_acceptable_kmp __P((struct cf_list *, void *));
static int rcf_fix_ikev1 __P((struct cf_list *, void *));
static int rcf_fix_ikev2 __P((struct cf_list *, void *));
static int rcf_fix_kink __P((struct cf_list *, void *));
static int rcf_fix_selector_index __P((struct cf_list *, void *));
static int rcf_fix_logmode __P((struct cf_list *, void *));
static int rcf_fix_passive __P((struct cf_list *, void *));
static int rcf_fix_logfile __P((struct cf_list *, void *));
static int rcf_fix_peers_ipaddr __P((struct cf_list *, void *));
static int rcf_fix_peers_kmp_port __P((struct cf_list *, void *));
static int rcf_fix_verify_id __P((struct cf_list *, void *));
static int rcf_fix_verify_pubkey __P((struct cf_list *, void *));
static int rcf_fix_send_cert __P((struct cf_list *, void *));
static int rcf_fix_send_cert_req __P((struct cf_list *, void *));
static int rcf_fix_nonce_size __P((struct cf_list *, void *));
static int rcf_fix_initial_contact __P((struct cf_list *, void *));
static int rcf_fix_support_proxy __P((struct cf_list *, void *));
static int rcf_fix_my_id __P((struct cf_list *, void *));
static int rcf_fix_peers_id __P((struct cf_list *, void *));
static int rcf_fix_selector_check __P((struct cf_list *, void *));
static int rcf_fix_proposal_check __P((struct cf_list *, void *));
static int rcf_fix_random_pad_content __P((struct cf_list *, void *));
static int rcf_fix_random_padlen __P((struct cf_list *, void *));
static int rcf_fix_max_padlen __P((struct cf_list *, void *));
static int rcf_fix_max_retry_to_send __P((struct cf_list *, void *));
static int rcf_fix_interval_to_send __P((struct cf_list *, void *));
static int rcf_fix_times_per_send __P((struct cf_list *, void *));
static int rcf_fix_kmp_sa_lifetime_time __P((struct cf_list *, void *));
static int rcf_fix_kmp_sa_lifetime_byte __P((struct cf_list *, void *));
static int rcf_fix_kmp_sa_nego_time_limit __P((struct cf_list *, void *));
static int rcf_fix_ipsec_sa_nego_time_limit __P((struct cf_list *, void *));
static int rcf_fix_kmp_enc_alg __P((struct cf_list *, void *));
static int rcf_fix_kmp_hash_alg __P((struct cf_list *, void *));
static int rcf_fix_kmp_prf_alg __P((struct cf_list *, void *));
static int rcf_fix_kmp_auth_method __P((struct cf_list *, void *));
static int rcf_fix_kmp_dh_group __P((struct cf_list *, void *));
static int rcf_fix_exchange_mode __P((struct cf_list *, void *));
static int rcf_fix_my_gssapi_id __P((struct cf_list *, void *));
static int rcf_fix_cookie_required __P((struct cf_list *, void *));
static int rcf_fix_my_principal __P((struct cf_list *, void *));
static int rcf_fix_peers_principal __P((struct cf_list *, void *));
static int rcf_fix_need_pfs __P((struct cf_list *, void *));
static int rcf_fix_my_public_key __P((struct cf_list *, void *));
static int rcf_fix_peers_public_key __P((struct cf_list *, void *));
static int rcf_fix_pre_shared_key __P((struct cf_list *, void *));
	/* selector */
static int rcf_fix_selector __P((struct rcf_selector **));
static void rcf_clean_selector_list __P((struct rcf_selector *head));
static void rcf_clean_selector __P((struct rcf_selector *sl));
static struct rcf_selector *rcf_deepcopy_selector __P((struct rcf_selector *));
static int rcf_fix_selector_order __P((struct cf_list *, void *));
static int rcf_fix_direction __P((struct cf_list *, void *));
static int rcf_fix_srcaddr __P((struct cf_list *, void *));
static int rcf_fix_dstaddr __P((struct cf_list *, void *));
static int rcf_fix_upper_layer_protocol __P((struct cf_list *, void *));
static int rcf_fix_next_header_including __P((struct cf_list *, void *));
static int rcf_setproto __P((vchar_t *, int *));
static int rcf_fix_policy_index __P((struct cf_list *, void *));
	/* policy */
static int rcf_fix_policy __P((vchar_t *pl_index, struct rcf_policy **));
static void rcf_clean_policy __P((struct rcf_policy *pl));
static struct rcf_policy *rcf_deepcopy_policy __P((struct rcf_policy *));
static int rcf_fix_action __P((struct cf_list *, void *));
static int rcf_fix_remote_index __P((struct cf_list *, void *));
static int rcf_fix_ipsec_index __P((struct cf_list *, void *));
static int rcf_fix_my_sa_addr __P((struct cf_list *, void *));
static int rcf_fix_peers_sa_addr __P((struct cf_list *, void *));
static int rcf_fix_ipsec_level __P((struct cf_list *, void *));
	/* ipsec */
static int rcf_fix_ipsec __P((vchar_t *ips_index, struct rcf_ipsec **));
static void rcf_clean_ipsec_list __P((struct rcf_ipsec *head));
static void rcf_clean_ipsec __P((struct rcf_ipsec *head));
static struct rcf_ipsec *rcf_deepcopy_ipsec __P((struct rcf_ipsec *));
static int rcf_fix_ipsec_sa_lifetime_time __P((struct cf_list *, void *));
static int rcf_fix_ipsec_sa_lifetime_byte __P((struct cf_list *, void *));
static int rcf_fix_ext_sequence __P((struct cf_list *, void *));
static int rcf_fix_ipsec_mode __P((struct cf_list *, void *));
static int rcf_fix_sa_index __P((struct cf_list *, void *));
	/* sa */
static int rcf_fix_sa __P((vchar_t *sa_index, struct rcf_sa **));
static void rcf_clean_sa __P((struct rcf_sa *sa));
static struct rcf_sa *rcf_deepcopy_sa __P((struct rcf_sa *));
static int rcf_fix_sa_protocol __P((struct cf_list *, void *));
static int rcf_fix_esp_enc_alg __P((struct cf_list *, void *));
static int rcf_fix_esp_auth_alg __P((struct cf_list *, void *));
static int rcf_fix_ah_auth_alg __P((struct cf_list *, void *));
static int rcf_fix_ipcomp_alg __P((struct cf_list *, void *));
static int rcf_fix_spi __P((struct cf_list *, void *));
	/* default */
static int rcf_fix_default __P((struct rcf_default **));
static void rcf_clean_default_list __P((struct rcf_default *));

	/* list parsing */
static void rcf_clean_kmp __P((struct rcf_kmp *));
static struct rcf_kmp *rcf_deepcopy_kmp __P((struct rcf_kmp *));
static void rcf_clean_log __P((struct rc_log *));
static struct rc_log *rcf_deepcopy_log __P((struct rc_log *));
static int rcf_fix_addrlist __P((struct cf_list *, struct rc_addrlist **, int, int));
static void rcf_clean_addrlist __P((struct rc_addrlist *));
static struct rc_addrlist *rcf_deepcopy_addrlist __P((struct rc_addrlist *));
static int rcf_fix_idlist __P((struct cf_list *, struct rc_idlist **));
static void rcf_clean_idlist __P((struct rc_idlist *));
static struct rc_idlist *rcf_deepcopy_idlist __P((struct rc_idlist *));
static int rcf_fix_alglist __P((struct cf_list *, struct rc_alglist **));
static void rcf_clean_alglist __P((struct rc_alglist *));
static struct rc_alglist *rcf_deepcopy_alglist __P((struct rc_alglist *));
static int rcf_fix_pklist __P((struct cf_list *, struct rc_pklist **));
static void rcf_clean_pklist __P((struct rc_pklist *));
static struct rc_pklist *rcf_deepcopy_pklist __P((struct rc_pklist *));
static int rcf_fix_string __P((struct cf_list *, vchar_t **));
static int rcf_fix_value __P((struct cf_list *, rc_type *));
static int rcf_fix_boolean __P((struct cf_list *, rc_type *));
static int rcf_fix_number __P((struct cf_list *, int *));
static int rcf_check_cfd __P((struct cf_list *, rcf_tdir));
static int rcf_check_cft __P((struct cf_list *, rcf_t));
static struct cf_list *rcf_get_cf_policy __P((vchar_t *));
static struct cf_list *rcf_get_cf_ipsec __P((vchar_t *));
static struct cf_list *rcf_get_cf_sa __P((vchar_t *));

struct rcf_tdf_t {
	rcf_tdir type;
	int (*tdf) __P((struct cf_list *, void *));
} rcf_tdf[] = {
	/* interface */
	{ CFD_IF_IKE,			rcf_fix_if_ike, },
	{ CFD_IF_KINK,			rcf_fix_if_kink, },
	{ CFD_IF_SPMD,			rcf_fix_if_spmd, },
	/* resolver */
	{ CFD_NAMESERVER,		rcf_fix_nameserver, },
	{ CFD_DNS_QUERY,		rcf_fix_dns_query, },
	{ CFD_RESOLVER,			rcf_fix_resolver_enable, },
	/* remote */
	{ CFD_IKEV1,			rcf_fix_ikev1, },
	{ CFD_IKEV2,			rcf_fix_ikev2, },
	{ CFD_KINK,			rcf_fix_kink, },
	{ CFD_ACCEPTABLE_KMP,		rcf_fix_acceptable_kmp, },
	{ CFD_SELECTOR_INDEX,		rcf_fix_selector_index, },
	{ CFD_LOGMODE,			rcf_fix_logmode, },
	{ CFD_LOGFILE,			rcf_fix_logfile, },
	{ CFD_PASSIVE,			rcf_fix_passive, },
	{ CFD_PEERS_IPADDR,		rcf_fix_peers_ipaddr, },
	{ CFD_PEERS_KMP_PORT,		rcf_fix_peers_kmp_port, },
	{ CFD_VERIFY_ID,		rcf_fix_verify_id, },
	{ CFD_VERIFY_PUBKEY,		rcf_fix_verify_pubkey, },
	{ CFD_SEND_CERT,		rcf_fix_send_cert, },
	{ CFD_SEND_CERT_REQ,		rcf_fix_send_cert_req, },
	{ CFD_NONCE_SIZE,		rcf_fix_nonce_size, },
	{ CFD_INITIAL_CONTACT,		rcf_fix_initial_contact, },
	{ CFD_SUPPORT_PROXY,		rcf_fix_support_proxy, },
	{ CFD_MY_ID,			rcf_fix_my_id, },
	{ CFD_PEERS_ID,			rcf_fix_peers_id, },
	{ CFD_SELECTOR_CHECK,		rcf_fix_selector_check, },
	{ CFD_PROPOSAL_CHECK,		rcf_fix_proposal_check, },
	{ CFD_RANDOM_PAD_CONTENT,	rcf_fix_random_pad_content, },
	{ CFD_RANDOM_PADLEN,		rcf_fix_random_padlen, },
	{ CFD_MAX_PADLEN,		rcf_fix_max_padlen, },
	{ CFD_MAX_RETRY_TO_SEND,	rcf_fix_max_retry_to_send, },
	{ CFD_INTERVAL_TO_SEND,		rcf_fix_interval_to_send, },
	{ CFD_TIMES_PER_SEND,		rcf_fix_times_per_send, },
	{ CFD_KMP_SA_LIFETIME_TIME,	rcf_fix_kmp_sa_lifetime_time, },
	{ CFD_KMP_SA_LIFETIME_BYTE,	rcf_fix_kmp_sa_lifetime_byte, },
	{ CFD_KMP_SA_NEGO_TIME_LIMIT,	rcf_fix_kmp_sa_nego_time_limit, },
	{ CFD_IPSEC_SA_NEGO_TIME_LIMIT,	rcf_fix_ipsec_sa_nego_time_limit, },
	{ CFD_KMP_ENC_ALG,		rcf_fix_kmp_enc_alg, },
	{ CFD_KMP_HASH_ALG,		rcf_fix_kmp_hash_alg, },
	{ CFD_KMP_PRF_ALG,		rcf_fix_kmp_prf_alg, },
	{ CFD_KMP_AUTH_METHOD,		rcf_fix_kmp_auth_method, },
	{ CFD_KMP_DH_GROUP,		rcf_fix_kmp_dh_group, },
	{ CFD_EXCHANGE_MODE,		rcf_fix_exchange_mode, },
	{ CFD_MY_GSSAPI_ID,		rcf_fix_my_gssapi_id, },
	{ CFD_COOKIE_REQUIRED,		rcf_fix_cookie_required, },
	{ CFD_MY_PRINCIPAL,		rcf_fix_my_principal, },
	{ CFD_PEERS_PRINCIPAL,		rcf_fix_peers_principal, },
	{ CFD_NEED_PFS,			rcf_fix_need_pfs, },
	{ CFD_MY_PUBLIC_KEY,		rcf_fix_my_public_key, },
	{ CFD_PEERS_PUBLIC_KEY,		rcf_fix_peers_public_key, },
	{ CFD_PRE_SHARED_KEY,		rcf_fix_pre_shared_key, },
	/* selector */
	{ CFD_SELECTOR_ORDER,		rcf_fix_selector_order, },
	{ CFD_DIRECTION,		rcf_fix_direction, },
	{ CFD_SRCADDR,			rcf_fix_srcaddr, },
	{ CFD_DSTADDR,			rcf_fix_dstaddr, },
	{ CFD_UPPER_LAYER_PROTOCOL,	rcf_fix_upper_layer_protocol, },
	{ CFD_NEXT_HEADER_INCLUDING,	rcf_fix_next_header_including, },
	{ CFD_POLICY_INDEX,		rcf_fix_policy_index, },
	/* policy */
	{ CFD_ACTION,			rcf_fix_action, },
	{ CFD_REMOTE_INDEX,		rcf_fix_remote_index, },
	{ CFD_IPSEC_INDEX,		rcf_fix_ipsec_index, },
	{ CFD_MY_SA_ADDR,		rcf_fix_my_sa_addr, },
	{ CFD_PEERS_SA_ADDR,		rcf_fix_peers_sa_addr, },
	{ CFD_IPSEC_LEVEL,		rcf_fix_ipsec_level, },
	/* ipsec */
	{ CFD_IPSEC_SA_LIFETIME_TIME,	rcf_fix_ipsec_sa_lifetime_time, },
	{ CFD_IPSEC_SA_LIFETIME_BYTE,	rcf_fix_ipsec_sa_lifetime_byte, },
	{ CFD_EXT_SEQUENCE,		rcf_fix_ext_sequence, },
	{ CFD_IPSEC_MODE,		rcf_fix_ipsec_mode, },
	{ CFD_SA_INDEX,			rcf_fix_sa_index, },
	/* sa */
	{ CFD_SA_PROTOCOL,		rcf_fix_sa_protocol, },
	{ CFD_ESP_ENC_ALG,		rcf_fix_esp_enc_alg, },
	{ CFD_ESP_AUTH_ALG,		rcf_fix_esp_auth_alg, },
	{ CFD_AH_AUTH_ALG,		rcf_fix_ah_auth_alg, },
	{ CFD_IPCOMP_ALG,		rcf_fix_ipcomp_alg, },
	{ CFD_SPI,			rcf_fix_spi, },
};

static int (*rcf_get_tdf(rcf_tdir dir))()
{
	int i;

	for (i = 0; i < ARRAYLEN(rcf_tdf); i++) {
		if (rcf_tdf[i].type == dir)
			return rcf_tdf[i].tdf;
	}

	return (void *)0;
}


/*
 * fix the configuration
 */
int
rcf_read(const char *file, int flag)
{
	struct rcf_setval *new_setval;
	struct rcf_default *new_default;
	struct rcf_interface *new_interface;
	struct rcf_resolver *new_resolver;
	struct rcf_remote *new_remote;
	struct rcf_selector *new_selector;

	if (rcf_init(flag))
		return -1;

	if (rcf_parse(file))
		return -1;

	if (rcf_fix_setval(&new_setval) ||
	    rcf_fix_default(&new_default) ||
	    rcf_fix_interface(&new_interface) ||
	    rcf_fix_resolver(&new_resolver) ||
	    rcf_fix_remote(&new_remote) ||
	    rcf_fix_selector(&new_selector))
		return -1;

	rcf_clean_cf();
	rcf_clean();
	rcf_setval_head = new_setval;
	rcf_default_head = new_default;
	rcf_interface_head = new_interface;
	rcf_resolver_head = new_resolver;
	rcf_remote_head = new_remote;
	rcf_selector_head = new_selector;

	return 0;
}

/*
 * clean the configuration trees
 */
int
rcf_clean()
{
	rcf_clean_setval_list(rcf_setval_head);
	rcf_clean_default_list(rcf_default_head);
	rcf_clean_interface_list(rcf_interface_head);
	rcf_clean_resolver_list(rcf_resolver_head);
	rcf_clean_remote_list(rcf_remote_head);
	rcf_clean_selector_list(rcf_selector_head);

	rcf_setval_head = 0;
	rcf_default_head = 0;
	rcf_interface_head = 0;
	rcf_resolver_head = 0;
	rcf_remote_head = 0;
	rcf_selector_head = 0;

	return 0;
}


/*
 * fix setval definitions
 */
static int
rcf_fix_setval(struct rcf_setval **dst)
{
	struct rcf_setval *new_head = 0, *new, *p;
	struct cf_list *n;

	for (n = cf_lists->cf_setval_head; n; n = n->nexts) {
		if ((new = rc_calloc(1, sizeof(*new))) == NULL) {
			plog(PLOG_CRITICAL, PLOGLOC, NULL,
			    "no memory %s %d\n", n->file, n->lineno);
			goto err;
		}
		if (rcf_fix_string(n, &new->sym)) {
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "error in %s %d\n", n->file, n->lineno);
			goto err;
		}
		if (rcf_fix_string(n->nextp, &new->val)) {
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "error in %s %d\n", n->file, n->lineno);
			goto err;
		}
		for (p = new_head; p && p->next; p = p->next)
			;
		if (p)
			p->next = new;
		else
			new_head = new;
	}
	*dst = new_head;

	return 0;

    err:
	rcf_clean_setval(new);
	rcf_clean_setval_list(new_head);

	return -1;
}

static void
rcf_clean_setval_list(struct rcf_setval *head)
{
	struct rcf_setval *n, *next;

	for (n = head; n; n = next) {
		next = n->next;
		rcf_clean_setval(n);
	}
}

static void
rcf_clean_setval(struct rcf_setval *head)
{
	if (!head)
		return;
	vfree(head->sym);
	vfree(head->val);
	rc_free(head);
}


/*
 * fix interface definitions
 */
static int
rcf_fix_interface(struct rcf_interface **dst)
{
	struct rcf_interface *new;
	struct cf_list *n;

	if ((new = rc_calloc(1, sizeof(*new))) == NULL) {
		plog(PLOG_CRITICAL, PLOGLOC, NULL, "no memory\n");
		return -1;
	}
	for (n = cf_lists->cf_interface_head; n; n = n->nexts)
		RCF_CALL_TDF(n, new);
	*dst = new;

	return 0;

    err:
	rcf_clean_interface_list(new);

	return -1;
}

static void
rcf_clean_interface_list(struct rcf_interface *head)
{
	if (!head)
		return;
	rcf_clean_addrlist(head->ike);
	rcf_clean_addrlist(head->kink);
	rcf_clean_addrlist(head->spmd);
	rc_free(head);
}

static int
rcf_fix_if_ike(struct cf_list *head, void *dst0)
{
	struct rcf_interface *dst = (struct rcf_interface *)dst0;
	int flag = RCT_ADDR_INET | RCT_ADDR_FQDN | RCT_ADDR_MACRO;

	if (rcf_check_cfd(head, CFD_IF_IKE))
		return -1;
	if (dst->ike) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error interface ike already existed "
		    "in %s %d\n", head->file, head->lineno);
		return -1;
	}
	if (rcf_fix_addrlist(head->nextp, &dst->ike, RC_PORT_IKE, flag)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_if_kink(struct cf_list *head, void *dst0)
{
	struct rcf_interface *dst = (struct rcf_interface *)dst0;
	int flag = RCT_ADDR_INET | RCT_ADDR_FQDN | RCT_ADDR_MACRO;

	if (rcf_check_cfd(head, CFD_IF_KINK))
		return -1;
	if (dst->kink) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error interface kink already existed "
		    "in %s %d\n", head->file, head->lineno);
		return -1;
	}
	if (rcf_fix_addrlist(head->nextp, &dst->kink, RC_PORT_KINK, flag)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_if_spmd(struct cf_list *head, void *dst0)
{
	struct rcf_interface *dst = (struct rcf_interface *)dst0;
	int flag = RCT_ADDR_INET | RCT_ADDR_FILE;

	if (rcf_check_cfd(head, CFD_IF_SPMD))
		return -1;
	/* this function should be called once in parsing a config file */
	if (dst->spmd) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error interface spmd already existed "
		    "in %s %d\n", head->file, head->lineno);
		return -1;
	}
	if (rcf_fix_addrlist(head->nextp, &dst->spmd, RC_PORT_SPMD, flag)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}


/*
 * fix resolver definitions
 */
static int
rcf_fix_resolver(struct rcf_resolver **dst)
{
	struct rcf_resolver *new;
	struct cf_list *n;

	if ((new = rc_calloc(1, sizeof(*new))) == NULL) {
		plog(PLOG_CRITICAL, PLOGLOC, NULL, "no memory\n");
		return -1;
	}
	for (n = cf_lists->cf_resolver_head; n; n = n->nexts)
		RCF_CALL_TDF(n, new);
	*dst = new;

	return 0;

    err:
	rcf_clean_resolver_list(new);

	return -1;
}

static void
rcf_clean_resolver_list(struct rcf_resolver *n)
{
	if (!n)
		return;
	rcf_clean_addrlist(n->nameserver);
	rcf_clean_addrlist(n->dns_query);
	rc_free(n);
}

static int
rcf_fix_nameserver(struct cf_list *head, void *dst0)
{
	struct rcf_resolver *dst = (struct rcf_resolver *)dst0;
	int flag = RCT_ADDR_INET;

	if (rcf_check_cfd(head, CFD_NAMESERVER))
		return -1;
	/* this function should be called once in parsing a config file */
	if (dst->nameserver) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error nameserver already existed "
		    "in %s %d\n", head->file, head->lineno);
		return -1;
	}
	if (rcf_fix_addrlist(head->nextp, &dst->nameserver, RC_PORT_NS, flag)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_dns_query(struct cf_list *head, void *dst0)
{
	struct rcf_resolver *dst = (struct rcf_resolver *)dst0;
	int flag = RCT_ADDR_INET;

	if (rcf_check_cfd(head, CFD_DNS_QUERY))
		return -1;
	/* this function should be called once in parsing a config file */
	if (dst->dns_query) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error dns_query already existed "
		    "in %s %d\n", head->file, head->lineno);
		return -1;
	}
	if (rcf_fix_addrlist(head->nextp, &dst->dns_query, RC_PORT_NSQUERY, flag)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_resolver_enable(struct cf_list *head, void *dst0)
{
	struct rcf_resolver *dst = (struct rcf_resolver *)dst0;

	if (rcf_check_cfd(head, CFD_RESOLVER))
		return -1;
	if (rcf_fix_boolean(head->nextp, &dst->resolver_enable)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}


/*
 * fix remote definitions
 */
static int
rcf_fix_remote(struct rcf_remote **dst)
{
	struct rcf_remote *new_head = 0, *new, *p;
	struct cf_list *n, *m;

	for (n = cf_lists->cf_remote_head; n; n = n->nexts) {
		if ((new = rc_calloc(1, sizeof(*new))) == NULL) {
			plog(PLOG_CRITICAL, PLOGLOC, NULL,
			    "no memory %s %d\n", n->file, n->lineno);
			goto err;
		}
		if (rcf_fix_string(n, &new->rm_index)) {
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "error in %s %d\n", n->file, n->lineno);
			goto err;
		}
		for (m = n->nextp; m; m = m->nexts)
			RCF_CALL_TDF(m, new);
		for (p = new_head; p && p->next; p = p->next)
			;
		if (p)
			p->next = new;
		else
			new_head = new;
	}
	*dst = new_head;

	return 0;

    err:
	rcf_clean_remote(new);
	rcf_clean_remote_list(new_head);

	return -1;
}

static void
rcf_clean_remote_list(struct rcf_remote *head)
{
	struct rcf_remote *n, *next;

	for (n = head; n; n = next) {
		next = n->next;
		rcf_clean_remote(n);
	}
}

static void
rcf_clean_remote(struct rcf_remote *n)
{
	if (!n)
		return;
	vfree(n->rm_index);
	rcf_clean_kmp(n->ikev1);
	rcf_clean_kmp(n->ikev2);
	rcf_clean_kmp(n->kink);
	rc_free(n);
}

static struct rcf_remote *
rcf_deepcopy_remote(struct rcf_remote *src)
{
	struct rcf_remote *new;

	if ((new = rc_calloc(1, sizeof(*new))) == NULL)
		return 0;
	new->sl_index = src->sl_index;
	new->initiate_kmp = src->initiate_kmp;
	new->acceptable_kmp = src->acceptable_kmp;
	DEEPCOPY_VDUP(src->rm_index, new->rm_index);
	DEEPCOPY_KMP(src->ikev1, new->ikev1);
	DEEPCOPY_KMP(src->ikev2, new->ikev2);
	DEEPCOPY_KMP(src->kink, new->kink);

	return new;

    err:
	plog(PLOG_INTERR, PLOGLOC, NULL, "something error happened.\n");
	rcf_clean_remote(new);
	return 0;
}

static int
rcf_fix_acceptable_kmp(struct cf_list *head, void *dst0)
{
	struct rcf_remote *dst = (struct rcf_remote *)dst0;
	struct cf_list *n;

	if (rcf_check_cfd(head, CFD_ACCEPTABLE_KMP))
		return -1;
	for (n = head->nextp; n; n = n->nexts) {
		if (rcf_check_cft(n, CFT_VALUE))
			return -1;
		if (!dst->initiate_kmp)
			dst->initiate_kmp = n->d.val;
		switch (n->d.val) {
		case RCT_KMP_IKEV1:
			dst->acceptable_kmp |= RCF_ALLOW_IKEV1;
			break;
		case RCT_KMP_IKEV2:
			dst->acceptable_kmp |= RCF_ALLOW_IKEV2;
			break;
		case RCT_KMP_KINK:
			dst->acceptable_kmp |= RCF_ALLOW_KINK;
			break;
		default:
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "error unknown kmp type %d in %s %d\n",
			    n->d.val, n->file, n->lineno);
			return -1;
		}
	}

	return 0;
}

static int
rcf_fix_ikev1(struct cf_list *head, void *dst0)
{
	struct rcf_remote *dst = (struct rcf_remote *)dst0;
	struct cf_list *n;
	struct rcf_kmp *new;

	if (rcf_check_cfd(head, CFD_IKEV1))
		return -1;
	if ((new = rc_calloc(1, sizeof(*new))) == NULL) {
		plog(PLOG_CRITICAL, PLOGLOC, NULL, "no memory\n");
		return -1;
	}
	new->kmp_proto = RCT_KMP_IKEV1;
	for (n = head->nextp; n; n = n->nexts) {
		RCF_CALL_TDF(n, new);
	}
	dst->ikev1 = new;

	return 0;

    err:
	rcf_clean_kmp(new);

	return -1;
}

static int
rcf_fix_ikev2(struct cf_list *head, void *dst0)
{
	struct rcf_remote *dst = (struct rcf_remote *)dst0;
	struct cf_list *n;
	struct rcf_kmp *new;

	if (rcf_check_cfd(head, CFD_IKEV2))
		return -1;
	if ((new = rc_calloc(1, sizeof(*new))) == NULL) {
		plog(PLOG_CRITICAL, PLOGLOC, NULL, "no memory\n");
		return -1;
	}
	new->kmp_proto = RCT_KMP_IKEV2;
	for (n = head->nextp; n; n = n->nexts)
		RCF_CALL_TDF(n, new);
	dst->ikev2 = new;

	return 0;

    err:
	rcf_clean_kmp(new);

	return -1;
}

static int
rcf_fix_kink(struct cf_list *head, void *dst0)
{
	struct rcf_remote *dst = (struct rcf_remote *)dst0;
	struct cf_list *n;
	struct rcf_kmp *new;

	if (rcf_check_cfd(head, CFD_KINK))
		return -1;
	if ((new = rc_calloc(1, sizeof(*new))) == NULL) {
		plog(PLOG_CRITICAL, PLOGLOC, NULL, "no memory\n");
		return -1;
	}
	new->kmp_proto = RCT_KMP_KINK;
	for (n = head->nextp; n; n = n->nexts)
		RCF_CALL_TDF(n, new);
	dst->kink = new;

	return 0;

    err:
	rcf_clean_kmp(new);

	return -1;
}

static int
rcf_fix_selector_index(struct cf_list *head, void *dst0)
{
	struct rcf_remote *dst = (struct rcf_remote *)dst0;

	if (rcf_check_cfd(head, CFD_SELECTOR_INDEX))
		return -1;
	if (rcf_fix_number(head->nextp, &dst->sl_index)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_logmode(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_LOGMODE))
		return -1;
	if (!dst->plog) {
		if ((dst->plog = rc_calloc(1, sizeof(*dst->plog))) == NULL) {
			plog(PLOG_CRITICAL, PLOGLOC, NULL, "no memory\n");
			return -1;
		}
	}
	if (rcf_fix_value(head->nextp, &dst->plog->logmode)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_logfile(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_LOGFILE))
		return -1;
	if (!dst->plog) {
		if ((dst->plog = rc_calloc(1, sizeof(*dst->plog))) == NULL) {
			plog(PLOG_CRITICAL, PLOGLOC, NULL, "no memory\n");
			return -1;
		}
	}
	if (rcf_fix_string(head->nextp, &dst->plog->logfile)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_passive(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_PASSIVE))
		return -1;
	if (rcf_fix_boolean(head->nextp, &dst->passive)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_peers_ipaddr(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;
	int flag = RCT_ADDR_INET | RCT_ADDR_FQDN | RCT_ADDR_MACRO;

	if (rcf_check_cfd(head, CFD_PEERS_IPADDR))
		return -1;
	if (rcf_fix_addrlist(head->nextp, &dst->peers_ipaddr, RC_PORT_ANY, flag)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_peers_kmp_port(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_PEERS_KMP_PORT))
		return -1;
	if (rcf_fix_number(head->nextp, &dst->peers_kmp_port)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_verify_id(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_VERIFY_ID))
		return -1;
	if (rcf_fix_boolean(head->nextp, &dst->verify_id)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_verify_pubkey(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_VERIFY_PUBKEY))
		return -1;
	if (rcf_fix_boolean(head->nextp, &dst->verify_pubkey)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_send_cert(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_SEND_CERT))
		return -1;
	if (rcf_fix_boolean(head->nextp, &dst->send_cert)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_send_cert_req(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_SEND_CERT_REQ))
		return -1;
	if (rcf_fix_boolean(head->nextp, &dst->send_cert_req)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_nonce_size(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_NONCE_SIZE))
		return -1;
	if (rcf_fix_number(head->nextp, &dst->nonce_size)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_initial_contact(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_INITIAL_CONTACT))
		return -1;
	if (rcf_fix_boolean(head->nextp, &dst->initial_contact)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_support_proxy(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_SUPPORT_PROXY))
		return -1;
	if (rcf_fix_boolean(head->nextp, &dst->support_proxy)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_my_id(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_MY_ID))
		return -1;
	if (rcf_fix_idlist(head->nextp, &dst->my_id)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_peers_id(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_PEERS_ID))
		return -1;
	if (rcf_fix_idlist(head->nextp, &dst->peers_id)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_selector_check(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_SELECTOR_CHECK))
		return -1;
	if (rcf_fix_value(head->nextp, &dst->selector_check)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_proposal_check(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_PROPOSAL_CHECK))
		return -1;
	if (rcf_fix_value(head->nextp, &dst->proposal_check)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_random_pad_content(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_RANDOM_PAD_CONTENT))
		return -1;
	if (rcf_fix_boolean(head->nextp, &dst->random_pad_content)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_random_padlen(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_RANDOM_PADLEN))
		return -1;
	if (rcf_fix_boolean(head->nextp, &dst->random_padlen)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_max_padlen(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_MAX_PADLEN))
		return -1;
	if (rcf_fix_number(head->nextp, &dst->max_padlen)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_max_retry_to_send(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_MAX_RETRY_TO_SEND))
		return -1;
	if (rcf_fix_number(head->nextp, &dst->max_retry_to_send)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_interval_to_send(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_INTERVAL_TO_SEND))
		return -1;
	if (rcf_fix_number(head->nextp, &dst->interval_to_send)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_times_per_send(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_TIMES_PER_SEND))
		return -1;
	if (rcf_fix_number(head->nextp, &dst->times_per_send)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_kmp_sa_lifetime_time(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_KMP_SA_LIFETIME_TIME))
		return -1;
	if (rcf_fix_number(head->nextp, &dst->kmp_sa_lifetime_time)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_kmp_sa_lifetime_byte(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_KMP_SA_LIFETIME_BYTE))
		return -1;
	if (rcf_fix_number(head->nextp, &dst->kmp_sa_lifetime_byte)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_kmp_sa_nego_time_limit(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_KMP_SA_NEGO_TIME_LIMIT))
		return -1;
	if (rcf_fix_number(head->nextp, &dst->kmp_sa_nego_time_limit)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_ipsec_sa_nego_time_limit(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_IPSEC_SA_NEGO_TIME_LIMIT))
		return -1;
	if (rcf_fix_number(head->nextp, &dst->ipsec_sa_nego_time_limit)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_kmp_enc_alg(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_KMP_ENC_ALG))
		return -1;
	if (rcf_fix_alglist(head->nextp, &dst->kmp_enc_alg)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_kmp_hash_alg(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_KMP_HASH_ALG))
		return -1;
	if (rcf_fix_alglist(head->nextp, &dst->kmp_hash_alg)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_kmp_prf_alg(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_KMP_PRF_ALG))
		return -1;
	if (rcf_fix_alglist(head->nextp, &dst->kmp_prf_alg)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_kmp_auth_method(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_KMP_AUTH_METHOD))
		return -1;
	if (rcf_fix_alglist(head->nextp, &dst->kmp_auth_method)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_kmp_dh_group(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_KMP_DH_GROUP))
		return -1;
	if (rcf_fix_alglist(head->nextp, &dst->kmp_dh_group)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_exchange_mode(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_EXCHANGE_MODE))
		return -1;
	if (rcf_fix_value(head->nextp, &dst->exchange_mode)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_my_gssapi_id(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_MY_GSSAPI_ID))
		return -1;
	if (rcf_fix_string(head->nextp, &dst->my_gssapi_id)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_cookie_required(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_COOKIE_REQUIRED))
		return -1;
	if (rcf_fix_boolean(head->nextp, &dst->cookie_required)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_my_principal(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_MY_PRINCIPAL))
		return -1;
	if (rcf_fix_string(head->nextp, &dst->my_principal)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_peers_principal(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_PEERS_PRINCIPAL))
		return -1;
	if (rcf_fix_string(head->nextp, &dst->peers_principal)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_need_pfs(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_NEED_PFS))
		return -1;
	if (rcf_fix_boolean(head->nextp, &dst->need_pfs)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_my_public_key(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_MY_PUBLIC_KEY))
		return -1;
	if (rcf_fix_pklist(head->nextp, &dst->my_pubkey)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_peers_public_key(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_PEERS_PUBLIC_KEY))
		return -1;
	if (rcf_fix_pklist(head->nextp, &dst->peers_pubkey)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_pre_shared_key(struct cf_list *head, void *dst0)
{
	struct rcf_kmp *dst = (struct rcf_kmp *)dst0;

	if (rcf_check_cfd(head, CFD_PRE_SHARED_KEY))
		return -1;
	if (rcf_fix_string(head->nextp, &dst->pre_shared_key)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}


/*
 * fix select definitions
 */
static int
rcf_fix_selector(struct rcf_selector **dst)
{
	struct rcf_selector *new_head = 0, *new, *p;
	struct cf_list *n, *m;

	for (n = cf_lists->cf_selector_head; n; n = n->nexts) {
		if ((new = rc_calloc(1, sizeof(*new))) == NULL) {
			plog(PLOG_CRITICAL, PLOGLOC, NULL,
			    "no memory %s %d\n", n->file, n->lineno);
			goto err;
		}
		if (rcf_fix_number(n, &new->sl_index)) {
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "error in %s %d\n", n->file, n->lineno);
			goto err;
		}
		/* set the default */
		new->upper_layer_protocol = RC_PROTO_ANY;
		/* fix each directive */
		for (m = n->nextp; m; m = m->nexts)
			RCF_CALL_TDF(m, new);
		for (p = new_head; p && p->next; p = p->next)
			;
		if (p)
			p->next = new;
		else
			new_head = new;
	}
	*dst = new_head;

	return 0;

    err:
	rcf_clean_selector(new);
	rcf_clean_selector_list(new_head);

	return -1;
}

static void
rcf_clean_selector_list(struct rcf_selector *head)
{
	struct rcf_selector *n, *next;

	for (n = head; n; n = next) {
		next = n->next;
		rcf_clean_selector(n);
	}
}

static void
rcf_clean_selector(struct rcf_selector *n)
{
	if (!n)
		return;
	rcf_clean_addrlist(n->src);
	rcf_clean_addrlist(n->dst);
	rcf_clean_policy(n->pl);
	rc_free(n);
}

static int
rcf_fix_selector_order(struct cf_list *head, void *dst0)
{
	struct rcf_selector *dst = (struct rcf_selector *)dst0;

	if (rcf_check_cfd(head, CFD_SELECTOR_ORDER))
		return -1;
	if (rcf_fix_number(head->nextp, &dst->order)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static struct rcf_selector *
rcf_deepcopy_selector(struct rcf_selector *src)
{
	struct rcf_selector *new;

	if ((new = rc_calloc(1, sizeof(*new))) == NULL)
		return 0;
	new->sl_index = src->sl_index;
	new->order = src->order;
	new->direction = src->direction;
	new->upper_layer_protocol = src->upper_layer_protocol;
	new->next_header_including = src->next_header_including;
	DEEPCOPY_ADDRLIST(src->src, new->src);
	DEEPCOPY_ADDRLIST(src->dst, new->dst);
	DEEPCOPY_POLICY(src->pl, new->pl);

	return new;

    err:
	plog(PLOG_INTERR, PLOGLOC, NULL, "something error happened.\n");
	rcf_clean_selector(new);
	return 0;
}

static int
rcf_fix_direction(struct cf_list *head, void *dst0)
{
	struct rcf_selector *dst = (struct rcf_selector *)dst0;

	if (rcf_check_cfd(head, CFD_DIRECTION))
		return -1;
	if (rcf_fix_value(head->nextp, &dst->direction)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_srcaddr(struct cf_list *head, void *dst0)
{
	struct rcf_selector *dst = (struct rcf_selector *)dst0;
	int flag = RCT_ADDR_INET | RCT_ADDR_FQDN | RCT_ADDR_MACRO;

	if (rcf_check_cfd(head, CFD_SRCADDR))
		return -1;
	if (rcf_fix_addrlist(head->nextp, &dst->src, RC_PORT_ANY, flag)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_dstaddr(struct cf_list *head, void *dst0)
{
	struct rcf_selector *dst = (struct rcf_selector *)dst0;
	int flag = RCT_ADDR_INET | RCT_ADDR_FQDN | RCT_ADDR_MACRO;

	if (rcf_check_cfd(head, CFD_DSTADDR))
		return -1;
	if (rcf_fix_addrlist(head->nextp, &dst->dst, RC_PORT_ANY, flag)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_upper_layer_protocol(struct cf_list *head, void *dst0)
{
	struct rcf_selector *dst = (struct rcf_selector *)dst0;
	vchar_t *proto;

	if (rcf_check_cfd(head, CFD_UPPER_LAYER_PROTOCOL))
		return -1;
	if (rcf_fix_string(head->nextp, &proto)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}
	if (rcf_setproto(proto, &dst->upper_layer_protocol)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "unknown protocol %s in %s %d\n", head->file, head->lineno);
		vfree(proto);
		return -1;
	}

	vfree(proto);
	return 0;
}

static int
rcf_fix_next_header_including(struct cf_list *head, void *dst0)
{
	struct rcf_selector *dst = (struct rcf_selector *)dst0;
	vchar_t *proto;

	if (rcf_check_cfd(head, CFD_NEXT_HEADER_INCLUDING))
		return -1;
	if (rcf_fix_string(head->nextp, &proto)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}
	if (rcf_setproto(proto, &dst->next_header_including)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "unknown protocol %s in %s %d\n",
		    proto, head->file, head->lineno);
		vfree(proto);
		return -1;
	}

	vfree(proto);
	return 0;
}

int
rcf_setproto(vchar_t *proto, int *num)
{
	struct protoent *pe;
	const char *name;
	char *bp;

	name = vmem2str(proto);
	if (name[0] == '0' && name[1] == 'x')
		*num = strtol(name, &bp, 16);
	else    
		*num = strtol(name, &bp, 10);
	if (*bp == '\0')
		return 0;

	/* assuming it is a string */
	if (strcmp("any", name) == 0) {
		*num = RC_PROTO_ANY;
		return 0;
	}
	if ((pe = getprotobyname(name)) == NULL)
		return -1;
	*num = pe->p_proto;

	return 0;
}

static int
rcf_fix_policy_index(struct cf_list *head, void *dst0)
{
	struct rcf_selector *dst = (struct rcf_selector *)dst0;
	vchar_t *pl_index;

	if (rcf_check_cfd(head, CFD_POLICY_INDEX))
		return -1;
	if (rcf_fix_string(head->nextp, &pl_index)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}
	/* extended a policy */
	if (rcf_fix_policy(pl_index, &dst->pl)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error getting a policy in %s %d\n",
		    head->file, head->lineno);
		vfree(pl_index);
		return -1;
	}
	vfree(pl_index);

	return 0;
}


/*
 * fix policy definitions
 */
static int
rcf_fix_policy(vchar_t *pl_index, struct rcf_policy **dst0)
{
	struct rcf_policy *new;
	struct cf_list *head, *n;

	if ((head = rcf_get_cf_policy(pl_index)) == 0) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error no policy for %s\n", vmem2str(pl_index));
		return -1;
	}
	if ((new = rc_calloc(1, sizeof(*new))) == NULL) {
		plog(PLOG_CRITICAL, PLOGLOC, NULL,
		    "no memory in rcf_fix_policy\n");
		return -1;
	}
	if (rcf_fix_string(head, &new->pl_index)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}
	for (n = head->nextp; n; n = n->nexts)
		RCF_CALL_TDF(n, new);
	*dst0 = new;

	return 0;

    err:
	rcf_clean_policy(new);

	return -1;
}

static void
rcf_clean_policy(struct rcf_policy *n)
{
	if (!n)
		return;
	vfree(n->pl_index);
	vfree(n->rm_index);
	rcf_clean_addrlist(n->my_sa_addr);
	rcf_clean_addrlist(n->peers_sa_addr);
	rcf_clean_ipsec_list(n->ips);
	rc_free(n);
}

static struct rcf_policy *
rcf_deepcopy_policy(struct rcf_policy *src)
{
	struct rcf_policy *new;

	if ((new = rc_calloc(1, sizeof(*new))) == NULL)
		return 0;
	new->action = src->action;
	new->ipsec_level = src->ipsec_level;
	DEEPCOPY_VDUP(src->rm_index, new->rm_index);
	DEEPCOPY_VDUP(src->pl_index, new->pl_index);
	DEEPCOPY_ADDRLIST(src->my_sa_addr, new->my_sa_addr);
	DEEPCOPY_ADDRLIST(src->peers_sa_addr, new->peers_sa_addr);
	DEEPCOPY_IPSEC(src->ips, new->ips);

	return new;

    err:
	plog(PLOG_INTERR, PLOGLOC, NULL, "something error happened.\n");
	rcf_clean_policy(new);
	return 0;
}

static int
rcf_fix_action(struct cf_list *head, void *dst0)
{
	struct rcf_policy *dst = (struct rcf_policy *)dst0;

	if (rcf_check_cfd(head, CFD_ACTION))
		return -1;
	if (rcf_fix_value(head->nextp, &dst->action)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_remote_index(struct cf_list *head, void *dst0)
{
	struct rcf_policy *dst = (struct rcf_policy *)dst0;

	if (rcf_check_cfd(head, CFD_REMOTE_INDEX))
		return -1;
	if (rcf_fix_string(head->nextp, &dst->rm_index)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_ipsec_index(struct cf_list *head, void *dst0)
{
	struct rcf_policy *dst = (struct rcf_policy *)dst0;
	struct cf_list *n;
	vchar_t *ips_index;

	if (rcf_check_cfd(head, CFD_IPSEC_INDEX))
		return -1;
	for (n = head->nextp; n; n = n->nexts) {
		if (rcf_fix_string(n, &ips_index)) {
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "error in %s %d\n", n->file, n->lineno);
			return -1;
		}
		/* extended a policy */
		if (rcf_fix_ipsec(ips_index, &dst->ips)) {
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "error in get a ipsec %s %d\n",
			    n->file, n->lineno);
			vfree(ips_index);
			return -1;
		}
		vfree(ips_index);
	}

	return 0;
}

static int
rcf_fix_my_sa_addr(struct cf_list *head, void *dst0)
{
	struct rcf_policy *dst = (struct rcf_policy *)dst0;
	int flag = RCT_ADDR_INET | RCT_ADDR_FQDN | RCT_ADDR_MACRO;

	if (rcf_check_cfd(head, CFD_MY_SA_ADDR))
		return -1;
	if (rcf_fix_addrlist(head->nextp, &dst->my_sa_addr, RC_PORT_ANY, flag)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_peers_sa_addr(struct cf_list *head, void *dst0)
{
	struct rcf_policy *dst = (struct rcf_policy *)dst0;
	int flag = RCT_ADDR_INET | RCT_ADDR_FQDN | RCT_ADDR_MACRO;

	if (rcf_check_cfd(head, CFD_PEERS_SA_ADDR))
		return -1;
	if (rcf_fix_addrlist(head->nextp, &dst->peers_sa_addr, RC_PORT_ANY, flag)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_ipsec_level(struct cf_list *head, void *dst0)
{
	struct rcf_policy *dst = (struct rcf_policy *)dst0;

	if (rcf_check_cfd(head, CFD_IPSEC_LEVEL))
		return -1;
	if (rcf_fix_value(head->nextp, &dst->ipsec_level)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}


/*
 * fix ipsec definitions
 */
static int
rcf_fix_ipsec(vchar_t *ips_index, struct rcf_ipsec **dst0)
{
	struct rcf_ipsec *new, *p;
	struct cf_list *head, *n;

	if ((head = rcf_get_cf_ipsec(ips_index)) == 0) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error no ipsec for %s\n", vmem2str(ips_index));
		return -1;
	}
	if ((new = rc_calloc(1, sizeof(*new))) == NULL) {
		plog(PLOG_CRITICAL, PLOGLOC, NULL, "no memory\n");
		return -1;
	}
	if (rcf_fix_string(head, &new->ips_index)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}
	for (n = head->nextp; n; n = n->nexts)
		RCF_CALL_TDF(n, new);
	for (p = *dst0; p && p->next; p = p->next)
		;
	if (p)
		p->next = new;
	else
		*dst0 = new;

	return 0;

    err:
	rcf_clean_ipsec_list(new);

	return -1;
}

static void
rcf_clean_ipsec_list(struct rcf_ipsec *head)
{
	struct rcf_ipsec *n, *next;

	for (n = head; n; n = next) {
		next = n->next;
		rcf_clean_ipsec(n);
	}
}

static struct rcf_ipsec *
rcf_deepcopy_ipsec(struct rcf_ipsec *src)
{
	struct rcf_ipsec *new_head = 0, *new = 0, *p, *n;

	for (n = src; n; n = n->next) {
		if ((new = rc_calloc(1, sizeof(*new))) == NULL)
			goto err;
		new->ipsec_sa_lifetime_time = n->ipsec_sa_lifetime_time;
		new->ipsec_sa_lifetime_byte = n->ipsec_sa_lifetime_byte;
		new->ext_sequence = n->ext_sequence;
		new->ipsec_mode = n->ipsec_mode;
		DEEPCOPY_VDUP(n->ips_index, new->ips_index);
		DEEPCOPY_SA(n->sa_ah, new->sa_ah);
		DEEPCOPY_SA(n->sa_esp, new->sa_esp);
		DEEPCOPY_SA(n->sa_ipcomp, new->sa_ipcomp);
		for (p = new_head; p && p->next; p = p->next)
			;
		if (p)
			p->next = new;
		else
			new_head = new;
	}
	return new_head;

    err:
	plog(PLOG_INTERR, PLOGLOC, NULL, "something error happened.\n");
	rcf_clean_ipsec(new);
	rcf_clean_ipsec_list(new_head);
	return 0;
}

static void
rcf_clean_ipsec(struct rcf_ipsec *n)
{
	if (!n)
		return;
	vfree(n->ips_index);
	rcf_clean_sa(n->sa_ah);
	rcf_clean_sa(n->sa_esp);
	rcf_clean_sa(n->sa_ipcomp);
	rc_free(n);
}

static int
rcf_fix_ipsec_sa_lifetime_time(struct cf_list *head, void *dst0)
{
	struct rcf_ipsec *dst = (struct rcf_ipsec *)dst0;

	if (rcf_check_cfd(head, CFD_IPSEC_SA_LIFETIME_TIME))
		return -1;
	if (rcf_fix_number(head->nextp, &dst->ipsec_sa_lifetime_time)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_ipsec_sa_lifetime_byte(struct cf_list *head, void *dst0)
{
	struct rcf_ipsec *dst = (struct rcf_ipsec *)dst0;

	if (rcf_check_cfd(head, CFD_IPSEC_SA_LIFETIME_BYTE))
		return -1;
	if (rcf_fix_number(head->nextp, &dst->ipsec_sa_lifetime_byte)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_ext_sequence(struct cf_list *head, void *dst0)
{
	struct rcf_ipsec *dst = (struct rcf_ipsec *)dst0;

	if (rcf_check_cfd(head, CFD_EXT_SEQUENCE))
		return -1;
	if (rcf_fix_boolean(head->nextp, &dst->ext_sequence)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_ipsec_mode(struct cf_list *head, void *dst0)
{
	struct rcf_ipsec *dst = (struct rcf_ipsec *)dst0;

	if (rcf_check_cfd(head, CFD_IPSEC_MODE))
		return -1;
	if (rcf_fix_value(head->nextp, &dst->ipsec_mode)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_sa_index(struct cf_list *head, void *dst0)
{
	struct rcf_ipsec *dst = (struct rcf_ipsec *)dst0;
	struct cf_list *n;
	vchar_t *sa_index;
	struct rcf_sa *sa_tmp, **sa_dst;

	if (rcf_check_cfd(head, CFD_SA_INDEX))
		return -1;
	for (n = head->nextp; n; n = n->nexts) {
		if (rcf_fix_string(n, &sa_index)) {
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "error in %s %d\n", n->file, n->lineno);
			return -1;
		}
		/* extended a policy */
		if (rcf_fix_sa(sa_index, &sa_tmp)) {
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "error in get a sa %s %d\n",
			    n->file, n->lineno);
			vfree(sa_index);
			return -1;
		}
		vfree(sa_index);
		switch (sa_tmp->sa_protocol) {
		case RCT_SATYPE_ESP:
			sa_dst = &dst->sa_esp;
			break;
		case RCT_SATYPE_AH:
			sa_dst = &dst->sa_ah;
			break;
		case RCT_SATYPE_IPCOMP:
			sa_dst = &dst->sa_ipcomp;
			break;
		default:
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "error unknown sa_protocol %d in %s %d\n",
			    sa_tmp->sa_protocol, n->file, n->lineno);
			return -1;
		}
		if (*sa_dst) {
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "error multiple %d defined in %s %d\n",
			    sa_tmp->sa_protocol, n->file, n->lineno);
			return -1;
		}
		*sa_dst = sa_tmp;
	}

	return 0;
}


/*
 * fix sa definitions
 */
static int
rcf_fix_sa(vchar_t *sa_index, struct rcf_sa **dst0)
{
	struct rcf_sa *new;
	struct cf_list *head, *n;

	if ((head = rcf_get_cf_sa(sa_index)) == 0) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error no sa for %s\n", vmem2str(sa_index));
		return -1;
	}
	if ((new = rc_calloc(1, sizeof(*new))) == NULL) {
		plog(PLOG_CRITICAL, PLOGLOC, NULL, "no memory\n");
		goto err;
	}
	if (rcf_fix_string(head, &new->sa_index)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}
	for (n = head->nextp; n; n = n->nexts)
		RCF_CALL_TDF(n, new);
	*dst0 = new;

	return 0;

    err:
	rcf_clean_sa(new);

	return -1;
}

static void
rcf_clean_sa(struct rcf_sa *n)
{
	if (!n)
		return;
	vfree(n->sa_index);
	rcf_clean_alglist(n->enc_alg);
	rcf_clean_alglist(n->auth_alg);
	rcf_clean_alglist(n->comp_alg);
	rc_free(n);
}

static struct rcf_sa *
rcf_deepcopy_sa(struct rcf_sa *src)
{
	struct rcf_sa *new;

	if ((new = rc_calloc(1, sizeof(*new))) == NULL)
		return 0;
	new->sa_protocol = src->sa_protocol;
	new->spi = src->spi;
	DEEPCOPY_VDUP(src->sa_index, new->sa_index);
	DEEPCOPY_ALGLIST(src->enc_alg, new->enc_alg);
	DEEPCOPY_ALGLIST(src->auth_alg, new->auth_alg);
	DEEPCOPY_ALGLIST(src->comp_alg, new->comp_alg);

	return new;

    err:
	plog(PLOG_INTERR, PLOGLOC, NULL, "something error happened.\n");
	rcf_clean_sa(new);
	return 0;
}

static int
rcf_fix_sa_protocol(struct cf_list *head, void *dst0)
{
	struct rcf_sa *dst = (struct rcf_sa *)dst0;

	if (rcf_check_cfd(head, CFD_SA_PROTOCOL))
		return -1;
	if (rcf_fix_value(head->nextp, &dst->sa_protocol)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_esp_enc_alg(struct cf_list *head, void *dst0)
{
	struct rcf_sa *dst = (struct rcf_sa *)dst0;

	if (rcf_check_cfd(head, CFD_ESP_ENC_ALG))
		return -1;
	if (rcf_fix_alglist(head->nextp, &dst->enc_alg)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_esp_auth_alg(struct cf_list *head, void *dst0)
{
	struct rcf_sa *dst = (struct rcf_sa *)dst0;

	if (rcf_check_cfd(head, CFD_ESP_AUTH_ALG))
		return -1;
	if (rcf_fix_alglist(head->nextp, &dst->auth_alg)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_ah_auth_alg(struct cf_list *head, void *dst0)
{
	struct rcf_sa *dst = (struct rcf_sa *)dst0;

	if (rcf_check_cfd(head, CFD_AH_AUTH_ALG))
		return -1;
	if (rcf_fix_alglist(head->nextp, &dst->auth_alg)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_ipcomp_alg(struct cf_list *head, void *dst0)
{
	struct rcf_sa *dst = (struct rcf_sa *)dst0;

	if (rcf_check_cfd(head, CFD_IPCOMP_ALG))
		return -1;
	if (rcf_fix_alglist(head->nextp, &dst->comp_alg)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}

static int
rcf_fix_spi(struct cf_list *head, void *dst0)
{
	struct rcf_sa *dst = (struct rcf_sa *)dst0;

	if (rcf_check_cfd(head, CFD_SPI))
		return -1;
	if (rcf_fix_number(head->nextp, &dst->spi)) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error in %s %d\n", head->file, head->lineno);
		return -1;
	}

	return 0;
}


/*
 * fix default definitions
 */
static int
rcf_fix_default(struct rcf_default **dst)
{
	struct rcf_default *new;
	struct rcf_remote *r_new, *rp;
	struct rcf_ipsec *i_new, *ip;
	struct cf_list *n, *m;

	if ((new = rc_calloc(1, sizeof(*new))) == NULL) {
		plog(PLOG_CRITICAL, PLOGLOC, NULL, "no memory\n");
		return -1;
	}
	r_new = NULL;
	i_new = NULL;

	for (n = cf_lists->cf_default_head; n; n = n->nexts) {
		switch (n->d.dir) {
		case CFD_REMOTE:
			if (new->remote) {
				plog(PLOG_INTERR, PLOGLOC, NULL,
				    "error default remote already exists "
				    "in %s %d\n", n->file, n->lineno);
				goto err;
			}
			if ((r_new = rc_calloc(1, sizeof(*r_new))) == NULL) {
				plog(PLOG_CRITICAL, PLOGLOC, NULL,
				    "no memory %s %d\n", n->file, n->lineno);
				goto err;
			}
			for (m = n->nextp; m; m = m->nexts)
				RCF_CALL_TDF(m, r_new);
			for (rp = new->remote; rp && rp->next; rp = rp->next)
				;
			if (rp)
				rp->next = r_new;
			else
				new->remote = r_new;
			r_new = NULL;
			break;
		case CFD_IPSEC:
			if (new->ipsec) {
				plog(PLOG_INTERR, PLOGLOC, NULL,
				    "error default ipsec already exists "
				    "in %s %d\n", n->file, n->lineno);
				goto err;
			}
			if ((i_new = rc_calloc(1, sizeof(*i_new))) == NULL) {
				plog(PLOG_CRITICAL, PLOGLOC, NULL,
				    "no memory %s %d\n", n->file, n->lineno);
				goto err;
			}
			for (m = n->nextp; m; m = m->nexts)
				RCF_CALL_TDF(m, i_new);
			for (ip = new->ipsec; ip && ip->next; ip = ip->next)
				;
			if (ip)
				ip->next = i_new;
			else
				new->ipsec = i_new;
			i_new = NULL;
			break;
		default:
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "error in %s %d\n", n->file, n->lineno);
			goto err;
		}
	}
	*dst = new;

	return 0;

    err:
	rcf_clean_remote(r_new);
	rcf_clean_remote_list(new->remote);
	rcf_clean_ipsec(i_new);
	rcf_clean_ipsec_list(new->ipsec);
	rc_free(new);

	return -1;
}

static void
rcf_clean_default_list(struct rcf_default *head)
{
	if (!head)
		return;
	rcf_clean_remote_list(head->remote);
	rcf_clean_ipsec_list(head->ipsec);
	rc_free(head);
	return;
}


static void
rcf_clean_kmp(struct rcf_kmp *n)
{
	if (!n)
		return;
	rcf_clean_log(n->plog);
	rcf_clean_addrlist(n->peers_ipaddr);
	rcf_clean_idlist(n->my_id);
	rcf_clean_idlist(n->peers_id);
	rcf_clean_alglist(n->kmp_enc_alg);
	rcf_clean_alglist(n->kmp_hash_alg);
	rcf_clean_alglist(n->kmp_prf_alg);
	rcf_clean_alglist(n->kmp_dh_group);
	rcf_clean_alglist(n->kmp_auth_method);
	vfree(n->my_gssapi_id);
	rcf_clean_pklist(n->my_pubkey);
	rcf_clean_pklist(n->peers_pubkey);
	vfree(n->pre_shared_key);
	vfree(n->my_principal);
	vfree(n->peers_principal);
	rc_free(n);
}

static struct rcf_kmp *
rcf_deepcopy_kmp(struct rcf_kmp *src)
{
	struct rcf_kmp *new;

	if ((new = rc_calloc(1, sizeof(*new))) == NULL)
		return 0;
	new->kmp_proto = src->kmp_proto;
	new->passive = src->passive;
	new->verify_id = src->verify_id;
	new->verify_pubkey = src->verify_pubkey;
	new->send_cert = src->send_cert;
	new->send_cert_req = src->send_cert_req;
	new->nonce_size = src->nonce_size;
	new->initial_contact = src->initial_contact;
	new->support_proxy = src->support_proxy;
	new->selector_check = src->selector_check;
	new->proposal_check = src->proposal_check;
	new->random_pad_content = src->random_pad_content;
	new->random_padlen = src->random_padlen;
	new->max_padlen = src->max_padlen;
	new->max_retry_to_send = src->max_retry_to_send;
	new->interval_to_send = src->interval_to_send;
	new->times_per_send = src->times_per_send;
	new->kmp_sa_lifetime_time = src->kmp_sa_lifetime_time;
	new->kmp_sa_lifetime_byte = src->kmp_sa_lifetime_byte;
	new->kmp_sa_nego_time_limit = src->kmp_sa_nego_time_limit;
	new->ipsec_sa_nego_time_limit = src->ipsec_sa_nego_time_limit;
	new->exchange_mode = src->exchange_mode;
	new->cookie_required = src->cookie_required;
	new->need_pfs = src->need_pfs;
	DEEPCOPY_LOG(src->plog, new->plog);
	DEEPCOPY_VDUP(src->my_gssapi_id, new->my_gssapi_id);
	DEEPCOPY_VDUP(src->my_principal, new->my_principal);
	DEEPCOPY_VDUP(src->peers_principal, new->peers_principal);
	DEEPCOPY_VDUP(src->pre_shared_key, new->pre_shared_key);
	DEEPCOPY_PKLIST(src->my_pubkey, new->my_pubkey);
	DEEPCOPY_PKLIST(src->peers_pubkey, new->peers_pubkey);
	DEEPCOPY_ADDRLIST(src->peers_ipaddr, new->peers_ipaddr);
	DEEPCOPY_IDLIST(src->my_id, new->my_id);
	DEEPCOPY_IDLIST(src->peers_id, new->peers_id);
	DEEPCOPY_ALGLIST(src->kmp_enc_alg, new->kmp_enc_alg);
	DEEPCOPY_ALGLIST(src->kmp_hash_alg, new->kmp_hash_alg);
	DEEPCOPY_ALGLIST(src->kmp_prf_alg, new->kmp_prf_alg);
	DEEPCOPY_ALGLIST(src->kmp_dh_group, new->kmp_dh_group);
	DEEPCOPY_ALGLIST(src->kmp_auth_method, new->kmp_auth_method);

	return new;

    err:
	plog(PLOG_INTERR, PLOGLOC, NULL, "something error happened.\n");
	rcf_clean_kmp(new);
	return 0;
}

static void
rcf_clean_log(struct rc_log *n)
{
	if (!n)
		return;
	vfree(n->logfile);
	rc_free(n);
}

static struct rc_log *
rcf_deepcopy_log(struct rc_log *src)
{
	struct rc_log *new;

	if ((new = rc_calloc(1, sizeof(*new))) == NULL)
		return 0;
	new->logmode = src->logmode;
	DEEPCOPY_VDUP(src->logfile, new->logfile);

	return new;

    err:
	plog(PLOG_INTERR, PLOGLOC, NULL, "something error happened.\n");
	rcf_clean_log(new);
	return 0;
}

/*
 * if you want to allow a single type, set a corresponded rc_type to "fixed".
 * otherwise it must be set to zero.
 * "fixed" can be set:
 *   RCT_ADDR_INET
 *   RCT_ADDR_FQDN
 *   RCT_ADDR_MACRO
 *   RCT_ADDR_FILE
 *   mixed them
 * for example, in case of RCT_ADDR_FQDN, it allows the strings as a string
 * like FQDN.
 */
static int
rcf_fix_addrlist(struct cf_list *head, struct rc_addrlist **dst,
    int default_port, int fixed)
{
	struct rc_addrlist *new_head = 0, *new = 0, **lastap;
	struct cf_list *n, *m;
	vchar_t va;
	struct rc_addrlist *al = 0;
	char port[10];
	int nport;
	int error;

	if (!head) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error addr must exist at least\n");
		return -1;
	}
	lastap = &new_head;
	for (n = head; n; n = n->nexts) {
		if (n->type == CFT_STRING) {
			/* get the port */
			m = n->nextp;
			if (!m) {
				snprintf(port, sizeof(port), "%d",
				    default_port);
				nport = default_port;
			} else if (m->type == CFT_STRING) {
				if ((nport = rcs_getport(m->d.str)) == -1) {
					plog(PLOG_INTERR, PLOGLOC, NULL,
					    "invalid port number %s\n",
					    m->d.str);
					goto err;
				}
				snprintf(port, sizeof(port), "%d", nport);
			} else {
				plog(PLOG_INTERR, PLOGLOC, NULL,
				    "invalid port definition\n");
				goto err;
			}
			va.l = strlen(n->d.str);
			va.v = n->d.str;
			error = rcs_getaddrlist(n->d.str, port, RCT_ADDR_FQDN, &al);
			if (error) {
				plog(PLOG_INTERR, PLOGLOC, NULL,
				    "%s in %s %d\n",
				    gai_strerror(error),
				    n->file, n->lineno);
				goto err;
			}
			if (al->next) {
				plog(PLOG_INTERR, PLOGLOC, NULL,
				    "addrlist returned multiple entries "
				    "in %s %d\n", n->file, n->lineno);
				goto err;
			}
			new = al;
		} else if (n->type == CFT_VALUE && n->d.val == RCT_ADDR_FILE) {
			if (fixed && !(fixed & RCT_ADDR_FILE)) {
				plog(PLOG_INTERR, PLOGLOC, NULL,
				    "file does not allowed to use "
				    "in %s %d\n", n->file, n->lineno);
				goto err;
			}
			if ((new = rc_calloc(1, sizeof(*new))) == NULL)
				goto err;
			if (rcf_fix_string(n->nextp, &new->a.vstr)) {
				plog(PLOG_INTERR, PLOGLOC, NULL,
				    "error in %s %d\n", n->file, n->lineno);
				goto err;
			}
			new->type = RCT_ADDR_FILE;
		} else {
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "unexpected cftype %d in %s %d\n",
			    n->type, n->file, n->lineno);
			goto err;
		}
		*lastap = new;
		lastap = &new->next;
	}
	*dst = new_head;

	return 0;

    err:
	if (new)
		free(new);
	plog(PLOG_INTERR, PLOGLOC, NULL,
	    "error %s %d\n", head->file, head->lineno);
	rcf_clean_addrlist(new_head);

	return -1;
}

static void
rcf_clean_addrlist(struct rc_addrlist *head)
{
	rcs_free_addrlist(head);
}

static struct rc_addrlist *
rcf_deepcopy_addrlist(struct rc_addrlist *src)
{
	struct rc_addrlist *new_head = 0, *new = 0, *p, *n;

	for (n = src; n; n = n->next) {
		if ((new = rc_calloc(1, sizeof(*new))) == NULL)
			return 0;
		new->type = n->type;
		new->port = n->port;
		new->prefixlen = n->prefixlen;
		switch (n->type) {
		case RCT_ADDR_INET:
			if ((new->a.ipaddr = rcs_sadup(n->a.ipaddr)) == 0)
				goto err;
			break;
		case RCT_ADDR_FQDN:
		case RCT_ADDR_MACRO:
		case RCT_ADDR_FILE:
			if ((new->a.vstr = vdup(n->a.vstr)) == 0)
				goto err;
			break;
		default:
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "error unknown address type %d "
			    "in a addrlist\n", n->type);
			return 0;
		}
		for (p = new_head; p && p->next; p = p->next)
			;
		if (p)
			p->next = new;
		else
			new_head = new;
	}
	return new_head;

    err:
	plog(PLOG_INTERR, PLOGLOC, NULL, "something error happened.\n");
	rcf_clean_addrlist(new);
	return 0;
}

static int
rcf_fix_idlist(struct cf_list *head, struct rc_idlist **dst)
{
	struct rc_idlist *new_head = 0, *new, *p;
	struct cf_list *n, *m;

	if (!head) {
		plog(PLOG_INTERR, PLOGLOC, NULL, "error id must exist\n");
		return -1;
	}
	for (n = head; n; n = n->nexts) {
		if ((new = rc_calloc(1, sizeof(*new))) == NULL) {
			plog(PLOG_CRITICAL, PLOGLOC, NULL,
			    "no memory %s %d\n", n->file, n->lineno);
			goto err;
		}
		m = n;
		if (rcf_fix_value(m, &new->idtype)) {
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "error near %s %d\n", m->file, m->lineno);
			goto err;
		}
		m = m->nextp;
		if (rcf_fix_string(m, &new->id)) {
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "error near %s %d\n", m->file, m->lineno);
			goto err;
		}
		for (p = new_head; p && p->next; p = p->next)
			;
		if (p)
			p->next = new;
		else
			new_head = new;
	}
	*dst = new_head;

	return 0;

    err:
	rcf_clean_idlist(new_head);

	return -1;
}

static void
rcf_clean_idlist(struct rc_idlist *head)
{
	struct rc_idlist *n, *next;

	for (n = head; n; n = next) {
		next = n->next;
		vfree(n->id);
		rc_free(n);
	}
}

static struct rc_idlist *
rcf_deepcopy_idlist(struct rc_idlist *src)
{
	struct rc_idlist *new_head = 0, *new = 0, *p, *n;

	for (n = src; n; n = n->next) {
		if ((new = rc_calloc(1, sizeof(*new))) == NULL)
			return 0;
		new->idtype = n->idtype;
		if ((new->id = vdup(n->id)) == 0)
			goto err;
		for (p = new_head; p && p->next; p = p->next)
			;
		if (p)
			p->next = new;
		else
			new_head = new;
	}
	return new_head;

    err:
	plog(PLOG_INTERR, PLOGLOC, NULL, "something error happened.\n");
	rcf_clean_idlist(new);
	return 0;
}

static int
rcf_fix_alglist(struct cf_list *head, struct rc_alglist **dst)
{
	struct rc_alglist *new_head = 0, *new, *p;
	struct cf_list *n, *m;

	if (!head) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error algorithm must exist\n");
		return -1;
	}
	if (*dst)
	    plog(PLOG_INTWARN, PLOGLOC, NULL,
		 "duplicated definition\n");
	for (n = head; n; n = n->nexts) {
		if ((new = rc_calloc(1, sizeof(*new))) == NULL) {
			plog(PLOG_CRITICAL, PLOGLOC, NULL,
			    "no memory %s %d\n", n->file, n->lineno);
			goto err;
		}
		m = n;
		if (rcf_fix_value(m, &new->algtype)) {
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "error near %s %d\n", m->file, m->lineno);
			goto err;
		}
		m = n->nextp;
		if (m) {
			if (rcf_fix_number(m, &new->keylen)) {
				plog(PLOG_INTERR, PLOGLOC, NULL,
				    "error in %s %d\n", m->file, m->lineno);
				return -1;
			}
			m = m->nextp;
			if (m) {
				if (rcf_fix_string(m, &new->key)) {
					plog(PLOG_INTERR, PLOGLOC, NULL,
					    "error in %s %d\n",
					    m->file, m->lineno);
					goto err;
				}
			}
		}
		/* XXX fixed the key length */
		if (new->keylen == RC_KEYLEN_NONE) {
		}
		/* XXX fixed the key */
		for (p = new_head; p && p->next; p = p->next)
			;
		if (p)
			p->next = new;
		else
			new_head = new;
	}
	*dst = new_head;

	return 0;

    err:
	rcf_clean_alglist(new_head);

	return -1;
}

static struct rc_alglist *
rcf_deepcopy_alglist(struct rc_alglist *src)
{
	struct rc_alglist *new_head = 0, *new = 0, *p, *n;

	for (n = src; n; n = n->next) {
		if ((new = rc_calloc(1, sizeof(*new))) == NULL)
			return 0;
		new->algtype = n->algtype;
		new->keylen = n->keylen;
		DEEPCOPY_VDUP(n->key, new->key);
		for (p = new_head; p && p->next; p = p->next)
			;
		if (p)
			p->next = new;
		else
			new_head = new;
	}
	return new_head;

    err:
	plog(PLOG_INTERR, PLOGLOC, NULL, "something error happened.\n");
	rcf_clean_alglist(new);
	return 0;
}

static void
rcf_clean_alglist(struct rc_alglist *head)
{
	struct rc_alglist *n, *next;

	for (n = head; n; n = next) {
		next = n->next;
		vfree(n->key);
		rc_free(n);
	}
}

static int
rcf_fix_pklist(struct cf_list *head, struct rc_pklist **dst)
{
	struct rc_pklist *new_head = 0, *new, *p;
	struct cf_list *n, *m;

	if (!head) {
		plog(PLOG_INTERR, PLOGLOC, NULL, "error pubkey must exist\n");
		return -1;
	}
	for (n = head; n; n = n->nexts) {
		if ((new = rc_calloc(1, sizeof(*new))) == NULL) {
			plog(PLOG_CRITICAL, PLOGLOC, NULL,
			    "no memory %s %d\n", n->file, n->lineno);
			goto err;
		}
		m = n;
		if (rcf_fix_value(m, &new->pktype)) {
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "error near %s %d\n", m->file, m->lineno);
			goto err;
		}
		m = m->nextp;
		if (rcf_fix_string(m, &new->pubkey)) {
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "error near %s %d\n", m->file, m->lineno);
			goto err;
		}
		m = m->nextp;
		if (rcf_fix_string(m, &new->privkey)) {
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "error near %s %d\n", m->file, m->lineno);
			goto err;
		}
		for (p = new_head; p && p->next; p = p->next)
			;
		if (p)
			p->next = new;
		else
			new_head = new;
	}
	*dst = new_head;

	return 0;

    err:
	rcf_clean_pklist(new_head);

	return -1;
}

static void
rcf_clean_pklist(struct rc_pklist *head)
{
	struct rc_pklist *n, *next;

	for (n = head; n; n = next) {
		next = n->next;
		vfree(n->pubkey);
		vfree(n->privkey);
		rc_free(n);
	}
}

static struct rc_pklist *
rcf_deepcopy_pklist(struct rc_pklist *src)
{
	struct rc_pklist *new_head = 0, *new = 0, *p, *n;

	for (n = src; n; n = n->next) {
		if ((new = rc_calloc(1, sizeof(*new))) == NULL)
			return 0;
		new->pktype = n->pktype;
		if ((new->pubkey = vdup(n->pubkey)) == 0 ||
		    (new->privkey = vdup(n->privkey)) == 0)
			goto err;
		for (p = new_head; p && p->next; p = p->next)
			;
		if (p)
			p->next = new;
		else
			new_head = new;
	}
	return new_head;

    err:
	plog(PLOG_INTERR, PLOGLOC, NULL, "something error happened.\n");
	rcf_clean_pklist(new);
	return 0;
}

/*
 * get a '\0' terminated string from a cf_list
 */
static int
rcf_fix_string(struct cf_list *n, vchar_t **dst)
{
	int len;

	if (rcf_check_cft(n, CFT_STRING))
		return -1;
	len = strlen(n->d.str);
	if (((*dst) = vmalloc(len)) == NULL) {
		plog(PLOG_CRITICAL, PLOGLOC, NULL,
		    "no memory %s %d\n", n->file, n->lineno);
		return -1;
	}
	memcpy((*dst)->v, n->d.str, len);

	return 0;
}

/*
 * get a rc_type value from a cf_list
 */
static int
rcf_fix_value(struct cf_list *n, rc_type *value)
{
	if (rcf_check_cft(n, CFT_VALUE))
		return -1;
	*value = n->d.val;

	return 0;
}

/*
 * get a boolean from a cf_list
 */
static int
rcf_fix_boolean(struct cf_list *n, rc_type *boolean)
{
	if (rcf_check_cft(n, CFT_VALUE))
		return -1;
	switch (n->d.val) {
	case RCT_BOOL_ON:
		*boolean = RCT_BOOL_ON;
		break;
	case RCT_BOOL_OFF:
		*boolean = RCT_BOOL_OFF;
		break;
	default:
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error unknown boolean type %d in %s %d\n",
		    n->d.val, n->file, n->lineno);
		return -1;
	}

	return 0;
}

/*
 * get a number from a cf_list
 */
static int
rcf_fix_number(struct cf_list *n, int *num)
{
	if (rcf_check_cft(n, CFT_NUMBER))
		return -1;
	*num = n->d.num;

	return 0;
}

/*
 * check a directive in a cf_list
 */
static int
rcf_check_cfd(struct cf_list *n, rcf_tdir dir)
{
	if (rcf_check_cft(n, CFT_DIRECTIVE))
		return -1;
	if (n->d.dir != dir) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error cfd %d expected but %d in %s %d\n",
		    dir, n->d.dir, n->file, n->lineno);
		return -1;
	}
	return 0;
}

/*
 * check a rc_type in a cf_list
 */
static int
rcf_check_cft(struct cf_list *n, rcf_t type)
{
	if (!n) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error null pointer passed in checking CFT\n");
		return -1;
	}
	if (n->type != type) {
		plog(PLOG_INTERR, PLOGLOC, NULL,
		    "error cft %d expected but %d in %s %d\n",
		    type, n->type, n->file, n->lineno);
		return -1;
	}
	return 0;
}

static struct cf_list *
rcf_get_cf_policy(vchar_t *pl_index)
{
	struct cf_list *n;
	vchar_t *index;

	for (n = cf_lists->cf_policy_head; n; n = n->nexts) {
		if (rcf_fix_string(n, &index)) {
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "error in %s %d\n", n->file, n->lineno);
			return 0;
		}
		if (vmemcmp(pl_index, index) == 0) {
			vfree(index);
			return n;
		}
		vfree(index);
	}

	return 0;
}

static struct cf_list *
rcf_get_cf_ipsec(vchar_t *ips_index)
{
	struct cf_list *n;
	vchar_t *index;

	for (n = cf_lists->cf_ipsec_head; n; n = n->nexts) {
		if (rcf_fix_string(n, &index)) {
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "error in %s %d\n", n->file, n->lineno);
			return 0;
		}
		if (vmemcmp(ips_index, index) == 0) {
			vfree(index);
			return n;
		}
		vfree(index);
	}

	return 0;
}

static struct cf_list *
rcf_get_cf_sa(vchar_t *sa_index)
{
	struct cf_list *n;
	vchar_t *index;

	for (n = cf_lists->cf_sa_head; n; n = n->nexts) {
		if (rcf_fix_string(n, &index)) {
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "error in %s %d\n", n->file, n->lineno);
			return 0;
		}
		if (vmemcmp(sa_index, index) == 0) {
			vfree(index);
			return n;
		}
		vfree(index);
	}

	return 0;
}


int
rcf_get_remotebyindex(vchar_t *rm_index, struct rcf_remote **dst)
{
	struct rcf_remote *src = 0, *n;

	for (n = rcf_remote_head; n; n = n->next) {
		if (vmemcmp(rm_index, n->rm_index) == 0) {
			src = n;
			break;
		}
	}
	if (!src)
		return -1;
	if (((*dst) = rcf_deepcopy_remote(src)) == 0)
		return -1;

	return 0;
}

int
rcf_get_remotebyaddr(struct sockaddr *s, rc_type proto, struct rcf_remote **dst)
{
	struct rcf_remote *src = 0, *n;
	struct rc_addrlist *al;
	struct rcf_kmp *kmp;

	for (n = rcf_remote_head; n; n = n->next) {
		switch (proto) {
		case RCT_KMP_IKEV1:
			kmp = n->ikev1;
			break;
		case RCT_KMP_IKEV2:
			kmp = n->ikev2;
			break;
		case RCT_KMP_KINK:
			kmp = n->kink;
			break;
		default:
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "invalid kmp type %d\n", proto);
			return -1;
		}
		if (!kmp || !kmp->peers_ipaddr)
			return -1;
		for (al = kmp->peers_ipaddr; al != 0; al = al->next) {
			if (al->type != RCT_ADDR_INET)
				continue;
			if (rcs_cmpsa_wop(al->a.ipaddr, s) != 0)
				continue;
			src = n;
			goto found;
		}
	}
  found:
	if (!src)
		return -1;
	if (((*dst) = rcf_deepcopy_remote(src)) == 0)
		return -1;

	return 0;
}

int
rcf_get_remotebypeersid(rc_type id_type, vchar_t *id_val, rc_type proto,
    struct rcf_remote **dst)
{
	struct rcf_remote *src = 0, *n;
	struct rc_idlist *idp;
	struct rcf_kmp *kmp;

	for (n = rcf_remote_head; n; n = n->next) {
		switch (proto) {
		case RCT_KMP_IKEV1:
			kmp = n->ikev1;
			break;
		case RCT_KMP_IKEV2:
			kmp = n->ikev2;
			break;
		case RCT_KMP_KINK:
			kmp = n->kink;
			break;
		default:
			plog(PLOG_INTERR, PLOGLOC, NULL,
			    "invalid kmp type %d\n", proto);
			return -1;
		}
		if (!kmp || !kmp->peers_id)
			return -1;
		for (idp = kmp->peers_id; idp != 0; idp = idp->next) {
			if (idp->idtype != id_type)
				continue;
			if (vmemcmp(idp->id, id_val) != 0)
				continue;
			src = n;
			goto found;
		}
	}
  found:
	if (!src)
		return -1;
	if (((*dst) = rcf_deepcopy_remote(src)) == 0)
		return -1;

	return 0;
}

void
rcf_free_remote(struct rcf_remote *n)
{
	rcf_clean_remote(n);
}

int
rcf_get_selectorlist(struct rcf_selector **dst)
{
	struct rcf_selector *new_head = 0, *new, *p, *n;

	for (n = rcf_selector_head; n; n = n->next) {
		if ((new = rcf_deepcopy_selector(n)) == 0) {
			rcf_clean_selector(new_head);
			return -1;
		}
		for (p = new_head; p && p->next; p = p->next)
			;
		if (p)
			p->next = new;
		else
			new_head = new;
	}
	*dst = new_head;

	return 0;
}

int
rcf_get_selector(int sl_index, struct rcf_selector **dst)
{
	struct rcf_selector *src = 0, *n;

	for (n = rcf_selector_head; n; n = n->next) {
		if (sl_index == n->sl_index) {
			src = n;
			break;
		}
	}
	if (!src)
		return -1;
	if (((*dst) = rcf_deepcopy_selector(src)) == 0)
		return -1;

	return 0;
}

void
rcf_free_selector(struct rcf_selector *n)
{
	rcf_clean_selector(n);
}

int
rcf_get_resolvers(struct rc_addrlist **dst)
{
	if (((*dst) = rcf_deepcopy_addrlist(rcf_resolver_head->nameserver))
	    == 0)
		return -1;
	
	return 0;
}

int
rcf_get_dns_queries(struct rc_addrlist **dst)
{
	if (((*dst) = rcf_deepcopy_addrlist(rcf_resolver_head->dns_query))
	    == 0)
		return -1;
	
	return 0;
}

int
rcf_spmd_resolver()
{
	if (!rcf_resolver_head)
		return 0;
	return rcf_resolver_head->resolver_enable;
}

int
rcf_get_spmd_interfaces(struct rc_addrlist **dst)
{
	if (((*dst) = rcf_deepcopy_addrlist(rcf_interface_head->spmd)) == 0)
		return -1;
	
	return 0;
}
