/* $Id: if_pfkeyv2.h,v 1.8 2005/01/07 09:02:09 sakane Exp $ */

/*
 * Copyright (C) 2003 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/socket.h>

#define RCPFK_ERRSTRBUFSIZE	128
#define RCPFK_SOCKBUFSIZE	128 * 1024

/* racoon PF_KEY message contener */
struct rcpfk_msg {
	int so;				/* pfkey socket */
	char estr[RCPFK_ERRSTRBUFSIZE];
	int eno;

	u_int32_t seq;
	struct sockaddr *sa_src;
	struct sockaddr *sa_dst;
	u_int32_t spi;
	u_int32_t reqid;
	u_int8_t satype;
	u_int8_t samode;
	u_int8_t enctype;
	u_int8_t authtype;
	size_t enckeylen;
	size_t authkeylen;
	caddr_t enckey;
	caddr_t authkey;
	u_int64_t lft_hard_time;
	u_int64_t lft_hard_bytes;
	u_int64_t lft_soft_time;
	u_int64_t lft_soft_bytes;
	u_int8_t expired;		/* 2:hard 1:soft */
	u_int8_t wsize;
	u_int32_t saflags;
	u_int32_t flags;
#define PFK_FLAG_NOHARM		0x00000001
#define PFK_FLAG_DEBUG		0x00000002

	u_int32_t slid;			/* "spid" in KAME impl. */
	struct sockaddr *sp_src;
	struct sockaddr *sp_dst;
	u_int8_t pref_src;
	u_int8_t pref_dst;
	u_int8_t ul_proto;
	u_int8_t dir;

	/* internal buffers; no need to touch from external */
	struct sockaddr_storage sa_src_storage;
	struct sockaddr_storage sa_dst_storage;
	struct sockaddr_storage sp_src_storage;
	struct sockaddr_storage sp_dst_storage;
};

struct rcpfk_cb {
	int (*cb_getspi) __P((struct rcpfk_msg *));
	int (*cb_update) __P((struct rcpfk_msg *));
	int (*cb_add) __P((struct rcpfk_msg *));
	int (*cb_expire) __P((struct rcpfk_msg *));
	int (*cb_acquire) __P((struct rcpfk_msg *));
	int (*cb_delete) __P((struct rcpfk_msg *));
	int (*cb_spdupdate) __P((struct rcpfk_msg *));
	int (*cb_spdadd) __P((struct rcpfk_msg *));
	int (*cb_spddelete) __P((struct rcpfk_msg *));
	int (*cb_spddelete2) __P((struct rcpfk_msg *));
	int (*cb_spdexpire) __P((struct rcpfk_msg *));
	int (*cb_spdget) __P((struct rcpfk_msg *));
};

extern int rcpfk_handler __P((struct rcpfk_msg *));
extern int rcpfk_init __P((struct rcpfk_msg *, struct rcpfk_cb *));
extern int rcpfk_clean __P((struct rcpfk_msg *));
extern int rcpfk_send_getspi __P((struct rcpfk_msg *));
extern int rcpfk_send_update __P((struct rcpfk_msg *));
extern int rcpfk_send_add __P((struct rcpfk_msg *));
extern int rcpfk_send_delete __P((struct rcpfk_msg *));
extern int rcpfk_send_acquire __P((struct rcpfk_msg *));
extern int rcpfk_send_register __P((struct rcpfk_msg *));
extern int rcpfk_send_spdupdate __P((struct rcpfk_msg *));
extern int rcpfk_send_spdadd __P((struct rcpfk_msg *));
extern int rcpfk_send_spddelete __P((struct rcpfk_msg *));
extern int rcpfk_send_spddelete2 __P((struct rcpfk_msg *));
extern int rcpfk_supported_auth __P((int));
extern int rcpfk_supported_enc __P((int));
