/*	$KAME: vmbuf.c,v 1.11 2001/11/26 16:54:29 sakane Exp $	*/

/*
 * Copyright (C) 2004 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/param.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "rc_malloc.h"
#include "vmbuf.h"
#include "rbuf.h"

vchar_t *
vmalloc(size)
	size_t size;
{
	vchar_t *var;

	if ((var = (vchar_t *)rc_malloc(sizeof(*var))) == NULL)
		return NULL;

	var->l = size;
	var->v = (caddr_t)rc_calloc(1, size);
	if (var->v == NULL) {
		(void)rc_free(var);
		return NULL;
	}

	return var;
}

vchar_t *
vrealloc(ptr, size)
	vchar_t *ptr;
	size_t size;
{
	caddr_t v;

	if (ptr != NULL) {
		if ((v = (caddr_t)rc_realloc(ptr->v, size)) == NULL)
			return NULL;
		if (size > ptr->l)
		    memset(v + ptr->l, 0, size - ptr->l);
		ptr->v = v;
		ptr->l = size;
	} else {
		if ((ptr = vmalloc(size)) == NULL)
			return NULL;
	}

	return ptr;
}

vchar_t *
vreallocf(ptr, size)
	vchar_t *ptr;
	size_t size;
{
	caddr_t v;

	if (ptr != NULL) {
		if ((v = (caddr_t)rc_realloc(ptr->v, size)) == NULL) {
			(void)vfree(ptr);
			return NULL;
		}
		memset(v + ptr->l, 0, size - ptr->l);
		ptr->v = v;
		ptr->l = size;
	} else {
		if ((ptr = vmalloc(size)) == NULL)
			return NULL;
	}

	return ptr;
}

void
vfree(var)
	vchar_t *var;
{
	if (var == NULL)
		return;

	if (var->v)
		(void)rc_free(var->v);

	(void)rc_free(var);

	return;
}

vchar_t *
vdup(src)
	const vchar_t *src;
{
	vchar_t *new;

	if ((new = vmalloc(src->l)) == NULL)
		return NULL;

	memcpy(new->v, src->v, src->l);

	return new;
}

int
vmemcmp(s1, s2)
        const vchar_t *s1, *s2;
{
	unsigned char *p1, *p2;
	int n;

	n = s1->l < s2->l ? s1->l : s2->l;
	if (n != 0) {
		p1 = s1->v;
		p2 = s2->v;
		do {
			if (*p1++ != *p2++)
				return (*--p1 - *--p2);
		} while (--n != 0);
	}
	return s1->l - s2->l;
}

/*
 * assumed that "src" must be a printable string.
 */
const char *
vmem2str(src)
	const vchar_t *src;
{
	vchar_t *buf;

	buf = rbuf_getvb(1 + src->l);
	memcpy(buf->v, src->v, src->l);
	buf->v[src->l] = '\0';

	return buf->v;
}

/*
 * copy a null terminated string to a allocated vmbuf.
 */
vchar_t *
str2vmem(src)
	const char *src;
{
	vchar_t *dst;
	int len;

	len = strlen(src);
	if ((dst = vmalloc(len)) == NULL)
		return NULL;
	memcpy(dst->v, src, len);

	return dst;
}

vchar_t *
vnew(ptr, len)
	const void *ptr;
	size_t len;
{
	vchar_t	* buf;

	buf = vmalloc(len);
	if (! buf) return 0;

	memcpy(buf->v, ptr, len);
	return buf;
}

vchar_t *
vprepend(buf, ptr, len)
	const vchar_t *buf;
	const void * ptr;
	size_t len;
{
	size_t orig_l = buf->l;
	vchar_t	* newv;

	if (! buf) return 0;
	newv = vmalloc(orig_l + len);
	if (! newv) return 0;
	memcpy(newv->v, ptr, len);
	memcpy(newv->v + len, buf->v, buf->l);
	return newv;
}
