/* $Id: utils.c,v 1.11 2004/12/17 09:03:39 mk Exp $ */
/*
 * Copyright (C) 2003 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "spmd_includes.h"

int spmd_loglevel = SPMD_L_DEFLT;

/* spmd_loglevel <-> syslog level table */
struct llv {
	int priority; /* spmd */
	int slg_lv;   /* syslog */
} llv_tbl [] = {
	{ SPMD_L_MIN, 0},
	{ SPMD_L_ERR, LOG_ERR, },
	{ SPMD_L_WARNING, LOG_WARNING, },
	{ SPMD_L_NOTICE, LOG_NOTICE, }, 
	{ SPMD_L_INFO, LOG_INFO, },
	{ SPMD_L_DEBUG, LOG_DEBUG, },
	{ SPMD_L_DEBUG2, LOG_DEBUG, },
	{ SPMD_L_MAX, 0},
};

void
spmd_log(int priority, const char *msg, ...)
{
	va_list args;
	char buf[256];

	if (priority > spmd_loglevel)
		return;

	va_start(args, msg);
	vsnprintf(buf, sizeof(buf), msg, args);
	va_end(args);

	if (spmd_foreground) {
		fprintf(stderr, "%s\n", buf);
	} else { 
		if (priority <= SPMD_L_MIN || priority >= SPMD_L_MAX)
			priority = SPMD_L_ERR;
		syslog(llv_tbl[priority].slg_lv, "%s", buf);
	}
	return;
}

int
sockcmp(struct sockaddr *sa1, struct sockaddr *sa2)
{

	if (sa1->sa_family != sa2->sa_family)
		return -1;

	if (sa1->sa_family == AF_INET) {
		struct sockaddr_in *sin1, *sin2;
		in_addr_t addr1, addr2;

		sin1 = (struct sockaddr_in *)sa1;
		sin2 = (struct sockaddr_in *)sa2;

		addr1 = sin1->sin_addr.s_addr;
		addr2 = sin2->sin_addr.s_addr;
		
		if (addr1 == addr2)
			return 0;
		else
			return 1;
	}

	if (sa1->sa_family == AF_INET6) {
		struct sockaddr_in6 *sin61, *sin62;
		struct in6_addr *addr61, *addr62;

		sin61 = (struct sockaddr_in6 *)sa1;
		sin62 = (struct sockaddr_in6 *)sa2;

		addr61 = &sin61->sin6_addr;
		addr62 = &sin62->sin6_addr;

		return (memcmp(addr61, addr62, sizeof(struct in6_addr)));

	}

	return -1; /* not reached */
}



/*---------- memory stuff ----------*/
memstat_t mstat[] =
{
	{M_EVER, 0, "MEM EVER"}, 
	{M_TMP, 0, "MEM TMP"},
	{M_QUERY, 0, "MEM QUERY"},
	{M_TASK, 0, "MEM TASK"},
	{M_DNS, 0, "MEM DNS"},
	{M_CACHE, 0, "MEM CACHE"},
	{M_RESOLV, 0, "MEM RESOLV"},
	{M_SHELL, 0, "MEM SHELL"},
	{M_END, 0, NULL},
};

void *
spmd_malloc(enum mtype type, size_t size)
{
	void *m;

	m = malloc(size);

	if (!m) {
		spmd_log(SPMD_L_WARNING, "%s:can not allocate memory",__FUNCTION__);
		goto fin;
	}

	mstat[type].allocated++;

fin:
	return m;
}

void *
spmd_calloc(enum mtype type, size_t size)
{

	void *m;

	m = calloc(1, size);

	if (!m) {
		spmd_log(SPMD_L_WARNING, "%s:can not allocate and set to zero",__FUNCTION__);
		goto fin;
	}

	mstat[type].allocated++;

fin:
	return m;
}

void
spmd_free(enum mtype type, void *p)
{
	free(p);

	mstat[type].allocated--;

	return;
}

char *
spmd_strdup(enum mtype type, const char *s)
{
	char *str;

	str = strdup(s);

	if (!str) {
		spmd_log(SPMD_L_WARNING, "%s:can not dup string",__FUNCTION__);
		goto fin;
	}

	mstat[type].allocated++;

fin:
	return str;
}


